"""Test cases of the _molecular_atmosphere module."""

import mitsuba as mi
import numpy as np
import numpy.testing as npt
import pytest

from eradiate import unit_registry as ureg
from eradiate.contexts import KernelContext
from eradiate.data import open_datasets
from eradiate.scenes.atmosphere import MolecularAtmosphere
from eradiate.scenes.core import Scene, traverse


def test_molecular_atmosphere_default_mono(mode_mono):
    atmosphere = MolecularAtmosphere()
    si = atmosphere.spectral_set().spectral_indices().__next__()
    template, _ = traverse(atmosphere)
    assert template.render(KernelContext(si=si))


def test_molecular_atmosphere_default_ckd(mode_ckd):
    atmosphere = MolecularAtmosphere()
    si = next(atmosphere.spectral_set().spectral_indices())
    template, _ = traverse(atmosphere)
    assert template.render(KernelContext(si=si))


def test_molecular_atmosphere_scale(mode_mono, error_handler_config):
    atmosphere = MolecularAtmosphere(
        thermoprops={
            "identifier": "afgl_1986-us_standard",
            "z": np.linspace(0, 120, 121) * ureg.km,
            "additional_molecules": False,
        },
        absorption_data=("komodo", [549.5, 550.5] * ureg.nm),
        error_handler_config=error_handler_config,
        scale=2.0,
    )
    template, _ = traverse(atmosphere)
    kernel_dict = template.render(KernelContext())
    assert kernel_dict["medium_atmosphere"]["scale"] == 2.0


def test_molecular_atmosphere_kernel_dict(mode_ckd, error_handler_config):
    """Constructor produces a valid kernel dictionary."""
    with open_datasets(
        [
            "spectra/absorption/ckd/monotropa/monotropa-35700_35800.nc",
            "spectra/absorption/ckd/monotropa/monotropa-9600_9700.nc",
            "spectra/absorption/ckd/monotropa/monotropa-18100_18200.nc",
            "spectra/absorption/ckd/monotropa/monotropa-4700_4800.nc",
        ]
    ) as absorption_data:
        atmosphere = MolecularAtmosphere(
            thermoprops={
                "identifier": "afgl_1986-us_standard",
                "z": np.linspace(0, 80, 41) * ureg.km,
                "additional_molecules": False,
            },
            absorption_data=absorption_data,
            error_handler_config=error_handler_config,
            geometry={
                "type": "spherical_shell",
                "ground_altitude": 0 * ureg.km,
                "toa_altitude": 80 * ureg.km,
            },
        )

        sis = list(atmosphere.spectral_set().spectral_indices())
        kernel_context = KernelContext(si=sis[0])

        template, params = traverse(Scene(objects={"atmosphere": atmosphere}))
        mi_scene: mi.Scene = mi.load_dict(template.render(kernel_context))

        # Mono evaluation generates valid parameter update tables
        mi_params: mi.SceneParameters = mi.traverse(mi_scene)

        # for w in eval_w:
        for si in sis:
            ctx = KernelContext(si=si)
            mi_params.update(params.render(ctx))


def test_molecular_atmosphere_switches(mode_mono, error_handler_config):
    # Absorption can be deactivated
    atmosphere = MolecularAtmosphere(
        absorption_data="spectra/absorption/mono/komodo/komodo.nc",
        has_absorption=False,
        error_handler_config=error_handler_config,
    )
    ctx = KernelContext()
    radprops = atmosphere.eval_radprops(ctx.si, optional_fields=True)
    npt.assert_allclose(radprops.sigma_a, 0.0)

    # Scattering can be deactivated
    atmosphere = MolecularAtmosphere(
        absorption_data="spectra/absorption/mono/komodo/komodo.nc",
        thermoprops={
            "identifier": "afgl_1986-us_standard",
            "z": np.linspace(0.0, 120.0, 121) * ureg.km,
            "additional_molecules": False,
        },
        has_scattering=False,
        error_handler_config=error_handler_config,
    )

    si = next(atmosphere.spectral_set().spectral_indices())
    radprops = atmosphere.eval_radprops(si, optional_fields=True)
    npt.assert_allclose(radprops.sigma_s, 0.0)

    # At least one must be active
    with pytest.raises(ValueError):
        MolecularAtmosphere(
            absorption_data="spectra/absorption/mono/komodo/komodo.nc",
            has_absorption=False,
            has_scattering=False,
            error_handler_config=error_handler_config,
        )



import mitsuba as mi

from eradiate.scenes.bsdfs import BlackBSDF
from eradiate.test_tools.types import check_scene_element


def test_black_constructor(modes_all):
    # Default constructor
    assert BlackBSDF()


def test_black_kernel_dict(modes_all_double):
    b = BlackBSDF()
    mi_wrapper = check_scene_element(
        b, mi.BSDF, drop_parameters=False
    )  # Do not drop untracked parameters: the reflectance (which we want to check) is untracked
    assert mi_wrapper.parameters["reflectance.value"] == 0.0



from pathlib import Path

import pytest

import eradiate
from eradiate.data import SafeDirectoryDataStore
from eradiate.exceptions import DataError

TEST_FILE_REGISTERED = Path("tests/data/git/registered_dataset.nc")
TEST_FILE_UNREGISTERED = Path("tests/data/git/unregistered_dataset.nc")


def test_directory_data_store_fetch():
    # The data submodule can be instantiated
    data_store = SafeDirectoryDataStore(
        path=eradiate.config.source_dir / "resources/data"
    )

    # We can fetch the test file
    assert data_store.fetch(TEST_FILE_REGISTERED)

    # Unregistered files cannot be fetched
    with pytest.raises(DataError):
        data_store.fetch(TEST_FILE_UNREGISTERED)



import numpy as np

from eradiate import warp


def check_inverse(func, inverse):
    for x in np.linspace(1e-6, 1 - 1e-6, 10):
        for y in np.linspace(1e-6, 1 - 1e-6, 10):
            p1 = np.array([x, y])
            p2 = func(p1)
            p3 = inverse(p2)
            assert np.allclose(p1, p3, atol=1e-5)


def test_square_to_uniform_disk_concentric():
    assert np.allclose(
        warp.square_to_uniform_disk_concentric([0, 0]),
        ([-1 / np.sqrt(2), -1 / np.sqrt(2)]),
    )
    assert np.allclose(warp.square_to_uniform_disk_concentric([0.5, 0.5]), [0, 0])

    check_inverse(
        warp.square_to_uniform_disk_concentric, warp.uniform_disk_to_square_concentric
    )


def test_square_to_uniform_hemisphere():
    assert np.allclose(warp.square_to_uniform_hemisphere([0.5, 0.5]), [0, 0, 1])
    assert np.allclose(warp.square_to_uniform_hemisphere([0, 0.5]), [-1, 0, 0])

    check_inverse(warp.square_to_uniform_hemisphere, warp.uniform_hemisphere_to_square)
    # check_vectorization("square_to_uniform_hemisphere")



import mitsuba as mi
import pytest

from eradiate.scenes.bsdfs import CheckerboardBSDF
from eradiate.test_tools.types import check_scene_element


@pytest.mark.parametrize(
    "kwargs",
    [{}, {"reflectance_a": 0.5, "reflectance_b": 0.1, "scale_pattern": 1.5}],
    ids=["noargs", "args"],
)
def test_checkerboard_construct(modes_all, kwargs):
    # Default constructor
    assert CheckerboardBSDF(**kwargs)


def test_checkerboard_kernel_dict(modes_all_double):
    bsdf = CheckerboardBSDF()
    mi_wrapper = check_scene_element(bsdf, mi.BSDF)
    assert "reflectance.color0.value" in mi_wrapper.parameters.keys()
    assert "reflectance.color1.value" in mi_wrapper.parameters.keys()



"""
Unit tests for the MultiDeltaSpectrum class.
"""

import numpy as np
import pint
import pytest

from eradiate import unit_registry as ureg
from eradiate.scenes.spectra import MultiDeltaSpectrum
from eradiate.spectral import BinSet


@pytest.mark.parametrize(
    "tested, expected",
    [
        ({}, TypeError),
        (
            {"wavelengths": [550.0, 660.0] * ureg.nm},
            MultiDeltaSpectrum(wavelengths=[550.0, 660.0] * ureg.nm),
        ),
        (
            {"wavelengths": [550.0, 660.0]},
            MultiDeltaSpectrum(wavelengths=[550.0, 660.0] * ureg.nm),
        ),
        (
            {"wavelengths": 550.0 * ureg.nm},
            MultiDeltaSpectrum(wavelengths=[550.0] * ureg.nm),
        ),
        ({"wavelengths": 550.0}, MultiDeltaSpectrum(wavelengths=[550.0] * ureg.nm)),
        ({"wavelengths": [550.0, -550.0]}, ValueError),
        (
            {"wavelengths": [660.0, 550.0]},
            MultiDeltaSpectrum(wavelengths=[550.0, 660.0] * ureg.nm),
        ),
        (
            {"wavelengths": [660.0, 550.0, 550.0]},
            MultiDeltaSpectrum(wavelengths=[550.0, 660.0] * ureg.nm),
        ),
    ],
    ids=[
        "no_args",
        "array_arg",
        "unitless_array_arg",
        "scalar_arg",
        "unitless_scalar_arg",
        "negative_values",
        "unsorted_values",
        "duplicate_values",
    ],
)
def test_multi_delta_construct(modes_all, tested, expected):
    if isinstance(expected, MultiDeltaSpectrum):
        s = MultiDeltaSpectrum(**tested)
        assert isinstance(s.wavelengths, pint.Quantity)
        assert np.allclose(s.wavelengths, expected.wavelengths)

    elif issubclass(expected, Exception):
        with pytest.raises(expected):
            MultiDeltaSpectrum(**tested)

    else:
        raise RuntimeError


@pytest.mark.parametrize(
    "multi_delta_wavelengths, expected_selected_wcenters",
    [
        ([555.0] * ureg.nm, [555.0] * ureg.nm),
        ([505.0, 595.0] * ureg.nm, [505.0, 595.0] * ureg.nm),
        (
            np.linspace(500.0, 600.0, 100) * ureg.nm,
            np.arange(505.0, 600.0, 10.0) * ureg.nm,
        ),
        ([510.0] * ureg.nm, [505.0] * ureg.nm),
    ],
    ids=[
        "single_value",
        "multiple_value",
        # If more values than bins are passed, still works as expected
        "many_values",
        # When wavelength falls between two bins, left bin is selected by convention
        "priority_left",
    ],
)
def test_multi_delta_select_ckd(multi_delta_wavelengths, expected_selected_wcenters):
    binset = BinSet.arange(start=500.0, stop=600.0, step=10.0)
    srf = MultiDeltaSpectrum(wavelengths=multi_delta_wavelengths)
    selected = srf.select_in(binset)
    assert np.allclose(selected.wcenters, expected_selected_wcenters)



from __future__ import annotations

import typing as t
import warnings

import matplotlib.pyplot as _plt
import numpy as np
import xarray.plot
from matplotlib.axes import Axes
from matplotlib.figure import Figure
from xarray.plot import FacetGrid


def set_style(rc=None):
    """
    Apply Eradiate Matplotlib style (the Seaborn ``ticks`` style with the
    viridis colormap).

    Parameters
    ----------
    rc : dict, optional
        A Matplotlib rc parameter dictionary to be applied in addition to the
        Eradiate style.
    """
    try:
        import seaborn
    except ModuleNotFoundError:
        warnings.warn(
            "To use Eradiate's Matplotlib style, you must install Seaborn.\n"
            "See instructions on https://seaborn.pydata.org/installing.html."
        )
        raise

    if rc is None:
        rc = {}

    seaborn.set_theme(style="ticks", rc={"image.cmap": "viridis", **rc})


def detect_axes(from_=None):
    """
    Try and extract a :class:`~matplotlib.axes.Axes` list from a data structure.

    Parameters
    ----------
    from_ : matplotlib figure or axes or list of axes or :class:`xarray.plot.FacetGrid`, optional
        Data structure to get an :class:`~matplotlib.axes.Axes` list from.
        If unset, :func:`matplotlib.pyplot.gca()` is used.

    Returns
    -------
    list of matplotlib axes
        Extracted list of :class:`~matplotlib.axes.Axes`.

    Raises
    ------
    TypeError
        If ``from_`` is of unsupported type.
    """
    if from_ is None:
        from_ = _plt.gca()

    if isinstance(from_, Figure):
        return from_.axes

    if isinstance(from_, Axes):
        return [from_]

    if isinstance(from_, FacetGrid):
        return list(from_.axes.flatten())

    if isinstance(from_, list):
        if all([isinstance(x, Axes) for x in from_]):
            return from_

    raise TypeError("unsupported type")


def get_axes_from_facet_grid(
    facet_grid: xarray.plot.FacetGrid, exclude: str = None
) -> list[Axes]:
    """
    Extract a flat list of :class:`~matplotlib.axes.Axes` from a
    :class:`~xarray.plot.FacetGrid`.

    Parameters
    ----------
    facet_grid : :class:`xarray.plot.FacetGrid`
        Object to extract a list of :class:`~matplotlib.axes.Axes`.

    exclude : str, optional
        Exclude selected :class:`~matplotlib.axes.Axes` from
        the returned list. Supported values:

        * ``"lower_left"``: exclude lower-left corner.

    Returns
    -------
    list of matplotlib axes
        List of extracted :class:`~matplotlib.axes.Axes` objects.
    """
    index = np.full((facet_grid.axes.size,), True, dtype=bool)
    n_rows, n_cols = facet_grid.axes.shape

    # This filtering system can be extended in the future
    if exclude == "lower_left":
        index[(n_rows - 1) * n_cols] = False

    return list(facet_grid.axes.flatten()[index])


def remove_xylabels(from_=None) -> None:
    """
    Remove x- and y-axis labels from ``from_``
    (processed by :func:`detect_axes`).

    Parameters
    ----------
    from_ : matplotlib figure or axes or list of axes or :class:`xarray.plot.FacetGrid`, optional
        Data structure to get an :class:`~matplotlib.axes.Axes` list from.
        If unset, :func:`matplotlib.pyplot.gca()` is used.

    See Also
    --------
    :func:`detect_axes`
    """
    for ax in detect_axes(from_):
        ax.set_xlabel("")
        ax.set_ylabel("")


def remove_xyticks(from_=None) -> None:
    """
    Remove x and y axis tick labels from ``from_``
    (processed by :func:`detect_axes`).

    Parameters
    ----------
    from_ : matplotlib figure or axes or list of axes or :class:`xarray.plot.FacetGrid`, optional
        Data structure to get an :class:`~matplotlib.axes.Axes` list from.
        If unset, :func:`matplotlib.pyplot.gca()` is used.

    See Also
    --------
    :func:`detect_axes`
    """
    for ax in detect_axes(from_):
        ax.get_xaxis().set_visible(False)
        ax.get_yaxis().set_visible(False)


def make_ticks(num_ticks: int, limits: t.Sequence[float]):
    """
    Generate ticks and their respective tickmarks.

    Parameters
    ----------
    num_ticks : int
        Number of ticks to generate, including the limits
        of the given range

    limits : tuple[float, float]
        List of two values, limiting the ticks inclusive

    Returns
    ------
    steps : list of float
        Values for the ticks.

    marks : list of str
        Tick values converted to degrees as string tickmarks.
    """

    step_width = float(limits[1] - limits[0]) / (num_ticks - 1)

    steps = [limits[0] + step_width * i for i in range(num_ticks)]
    marks = [f"{i / np.pi * 180}°" for i in steps]

    return steps, marks



import pathlib

import numpy as np
import pytest

from eradiate.kernel import read_binary_grid3d, write_binary_grid3d


@pytest.mark.parametrize(
    "shape",
    [
        (1, 1, 10),
        (3, 3, 4, 3),
        (4, 5, 1, 4),
    ],
)
def test_write_read_binary_grid3d(modes_all, shape, tmpdir):
    """Reads what was written."""

    length = np.prod(shape)
    write_values = np.random.random(length).reshape(shape)

    tmp_filename = pathlib.Path(tmpdir, "test.vol")
    write_binary_grid3d(filename=tmp_filename, values=write_values)
    read_values = read_binary_grid3d(tmp_filename)

    assert np.allclose(write_values, read_values)



import os
import time
from pathlib import Path

import pytest

from eradiate.data import BlindOnlineDataStore
from eradiate.exceptions import DataError

TEST_STORE = "http://eradiate.eu/data/store/stable"


def test_blind_data_store_fetch(tmpdir):
    test_file = Path("tests/data/online/unregistered_dataset.nc")

    # Using an empty storage directory, a BlindOnlineDataStore instance can be created
    store = BlindOnlineDataStore(base_url=TEST_STORE, path=tmpdir)

    # We can download a resource if it is available online
    fpath = store.fetch(test_file)
    assert fpath.is_file()

    # Downloaded files are reused if they exist
    last_modified = os.path.getmtime(fpath)
    os.remove(fpath)
    time.sleep(0.1)
    assert last_modified != os.path.getmtime(store.fetch(test_file))
    last_modified = os.path.getmtime(fpath)
    assert last_modified == os.path.getmtime(store.fetch(test_file))

    # Requesting missing resources raises
    with pytest.raises(DataError):
        store.fetch("foo")


@pytest.mark.parametrize("keep", [False, True])
def test_blind_data_store_purge(tmpdir, keep):
    store = BlindOnlineDataStore(base_url=TEST_STORE, path=tmpdir)

    files = [
        "tests/data/online/unregistered_dataset.nc.gz",
        "tests/data/online/registered_dataset.nc.gz",
    ]

    # Download files
    for file in files:
        store.fetch(file)
    assert len(list(os.scandir(tmpdir / "tests/data/online"))) == 2

    if not keep:
        store.purge(keep=None)
        # All files are removed
        assert len(list(os.scandir(tmpdir))) == 0

    else:
        store.purge(keep=["tests/data/online/unregistered_dataset.nc.gz"])
        # Specified files are kept, the others are removed
        assert len(list(os.scandir(tmpdir))) == 1
        assert (tmpdir / "tests/data/online/unregistered_dataset.nc.gz").isfile()
        assert not (tmpdir / "tests/data/online/registered_dataset.nc.gz").isfile()



import mitsuba as mi
import pytest

from eradiate.scenes.bsdfs import LambertianBSDF
from eradiate.scenes.spectra import UniformSpectrum
from eradiate.test_tools.types import check_scene_element


@pytest.mark.parametrize(
    "kwargs",
    [{}, {"reflectance": 0.5}],
    ids=["noargs", "args"],
)
def test_lambertian_construct(modes_all, kwargs):
    # Default constructor
    bsdf = LambertianBSDF(**kwargs)
    assert isinstance(bsdf.reflectance, UniformSpectrum)


def test_lambertian_kernel_dict(modes_all_double):
    bsdf = LambertianBSDF(reflectance=0.75)

    mi_wrapper = check_scene_element(bsdf, mi.BSDF)
    assert mi_wrapper.parameters["reflectance.value"] == 0.75



import numpy as np
import xarray as xr

from eradiate.xarray.interp import dataarray_to_rgb, film_to_angular


def test_film_to_angular():
    # This does not test the output very well, but it will fail if any
    # subcomponent is broken.
    # Actual behaviour is better tested manually by comparing film-space and
    # polar plots.

    # Create test data array
    da = xr.DataArray(
        [[i + j for i in range(16)] for j in range(32)],
        coords=(
            ("x", np.arange(0.5, 32, 1) / 32),
            ("y", np.arange(0.5, 16, 1) / 16),
        ),
    )

    theta = np.radians(np.arange(0.0, 85.0, 5))
    phi = np.radians(np.arange(0.0, 360.01, 5))

    da_angular = film_to_angular(da, theta=theta, phi=phi)
    assert set(da_angular.coords) == {"theta", "phi"}


def test_dataarray_to_rgb():
    r = 0.25 * np.ones((3, 2, 1))
    g = 0.50 * np.ones((3, 2, 1))
    b = 0.75 * np.ones((3, 2, 1))
    da = xr.DataArray(
        data=np.concatenate([b, g, r], axis=2),
        dims=["x", "y", "w"],
        coords={"x": [0, 1, 2], "y": [0, 1], "w": [440, 550, 660]},
    )

    channels = [("w", 660), ("w", 550), ("w", 440)]

    # Everything off: the data are just assembled in the same order as the channels
    img = dataarray_to_rgb(da, channels, normalize=False, gamma_correction=False)
    assert isinstance(img, np.ndarray)
    assert np.allclose(img, da.values[:, :, ::-1])

    # Normalisation scales the data
    img = dataarray_to_rgb(da, channels, normalize=True, gamma_correction=False)
    assert np.max(img) == 1.0

    # Gamma correction applies the gamma 2.2 operator
    img = dataarray_to_rgb(da, channels, normalize=False, gamma_correction=True)
    assert np.allclose(img, da.values[:, :, ::-1] ** (1.0 / 2.2))

    # By default, both are applied
    img = dataarray_to_rgb(da, channels)
    assert np.allclose(img, (da.values[:, :, ::-1] / float(da.max())) ** (1.0 / 2.2))



"""
Atmosphere's radiative profile.
"""
from __future__ import annotations

from copy import deepcopy

import attrs
import joseki
import numpy as np
import pint
import xarray as xr
from joseki.profiles.core import interp

from ._core import RadProfile, ZGrid, make_dataset
from .absorption import (
    DEFAULT_HANDLER_CONFIG,
    eval_sigma_a_ckd_impl,
    eval_sigma_a_mono_impl,
)
from .rayleigh import compute_sigma_s_air
from ..attrs import documented, parse_docs
from ..converters import convert_absorption_data, convert_thermoprops
from ..units import to_quantity
from ..units import unit_registry as ureg
from ..util.misc import cache_by_id, summary_repr
from ..validators import validate_absorption_data


def _absorption_data_repr(
    value: dict[tuple[pint.Quantity], xr.Dataset]
) -> dict[str, str]:
    def repr_k(value):
        """Representation for keys which are wavelength intervals."""
        units = value[0].u
        tmp = (value[0].m, value[1].m_as(units)) * units
        return f"{tmp:.2f~}"

    def repr_v(value):
        """Representation for values which are absorption dataset"""
        return summary_repr(value)

    if len(value) < 10:
        lines = [f"{repr_k(k)}: {repr_v(v)}" for k, v in value.items()]

    else:
        lines = []

        for i, (k, v) in enumerate(value.items()):
            if i > 4:
                break
            lines.append(f"{repr_k(k)}: {repr_v(v)}")

        lines.append("...")

        for i, (k, v) in enumerate(reversed(value.items())):
            if i > 4:
                break
            lines.append(f"{repr_k(k)}: {repr_v(v)}")

    return "\n".join(lines)


@parse_docs
@attrs.define(eq=False)
class AtmosphereRadProfile(RadProfile):
    """
    Atmospheric radiative profile.

    Notes
    -----
    The radiative profile is defined by atmospheric thermophysical and
    absorption coefficient data.
    """

    absorption_data: xr.Dataset | list[xr.Dataset] = documented(
        attrs.field(
            converter=convert_absorption_data,
            validator=validate_absorption_data,
            repr=_absorption_data_repr,
        ),
        doc="Absorption coefficient data. "
        "If a file path, the absorption coefficient is loaded from the "
        "specified file (must be a NetCDF file)."
        "If a tuple, the first element is the dataset codename and the"
        "second is the desired working wavelength range.",
        type="Dataset or list of Dataset",
        init_type="Dataset or list of Dataset or :class:`.PathLike`",
    )

    thermoprops: xr.Dataset = documented(
        attrs.field(
            factory=lambda: joseki.make(
                identifier="afgl_1986-us_standard",
                z=np.linspace(0.0, 120.0, 121) * ureg.km,
                additional_molecules=False,
            ),
            converter=convert_thermoprops,
            repr=summary_repr,
        ),
        doc="Atmosphere's thermophysical properties.",
        type="Dataset",
        default=(
            "`joseki.make <https://rayference.github.io/joseki/latest/reference/#src.joseki.core.make>`_"
            ' with ``identifier`` set to "afgl_1986-us_standard" and '
            '``z`` set to "np.linspace(0.0, 120.0, 121) * ureg.km" and '
            '``additional_molecules`` set to "False".'
        ),
    )

    @thermoprops.validator
    def _check_thermoprops(self, attribute, value):
        if not value.joseki.is_valid:
            raise ValueError(
                "Invalid thermophysical properties dataset."  # TODO: explain what is invalid
            )

    has_absorption: bool = documented(
        attrs.field(
            default=True,
            converter=bool,
            validator=attrs.validators.instance_of(bool),
        ),
        doc="Absorption switch. If ``True``, the absorption coefficient is "
        "computed. Else, the absorption coefficient is not computed and "
        "instead set to zero.",
        type="bool",
        default="True",
    )

    has_scattering: bool = documented(
        attrs.field(
            default=True,
            converter=bool,
            validator=attrs.validators.instance_of(bool),
        ),
        doc="Scattering switch. If ``True``, the scattering coefficient is "
        "computed. Else, the scattering coefficient is not computed and "
        "instead set to zero.",
        type="bool",
        default="True",
    )

    _zgrid: ZGrid | None = attrs.field(default=None, init=False)

    error_handler_config: dict[str, dict[str, str]] = documented(
        attrs.field(
            factory=lambda: deepcopy(DEFAULT_HANDLER_CONFIG),
            validator=attrs.validators.deep_mapping(
                key_validator=attrs.validators.instance_of(str),
                value_validator=attrs.validators.deep_mapping(
                    key_validator=attrs.validators.instance_of(str),
                    value_validator=attrs.validators.instance_of(str),
                ),
            ),
        ),
        doc="Error handler configuration for absorption data interpolation.",
        type="dict",
        default=DEFAULT_HANDLER_CONFIG,
    )

    def __attrs_post_init__(self):
        self.update()

    def update(self) -> None:
        self._zgrid = ZGrid(levels=self.levels)

    @property
    def levels(self) -> pint.Quantity:
        return to_quantity(self.thermoprops.z)

    @property
    def zgrid(self) -> ZGrid:
        # Inherit docstring
        return self._zgrid

    @cache_by_id
    def _thermoprops_interp(self, zgrid: ZGrid) -> xr.Dataset:
        # Interpolate thermophysical profile on specified altitude grid
        # Note: this value is cached so that repeated calls with the same zgrid
        #       won't trigger an unnecessary computation.
        return interp(
            self.thermoprops,
            z_new=zgrid.levels.m * ureg(str(zgrid.levels.units)),
            method={"default": "nearest"},  # TODO: revisit
        )

    def eval_albedo_mono(self, w: pint.Quantity, zgrid: ZGrid) -> pint.Quantity:
        # Inherit docstring
        sigma_s = self.eval_sigma_s_mono(w, zgrid)
        sigma_t = self.eval_sigma_t_mono(w, zgrid)
        return np.divide(
            sigma_s, sigma_t, where=sigma_t != 0.0, out=np.zeros_like(sigma_s)
        ).to("dimensionless")

    def eval_albedo_ckd(
        self,
        w: pint.Quantity,
        g: float,
        zgrid: ZGrid,
    ) -> pint.Quantity:
        sigma_s = self.eval_sigma_s_ckd(w=w, g=g, zgrid=zgrid)
        sigma_t = self.eval_sigma_t_ckd(w=w, g=g, zgrid=zgrid)
        return np.divide(
            sigma_s, sigma_t, where=sigma_t != 0.0, out=np.zeros_like(sigma_s)
        ).to("dimensionless")

    def eval_sigma_a_mono(self, w: pint.Quantity, zgrid: ZGrid) -> pint.Quantity:
        # NOTE: this method accepts 'w'-arrays and is vectorized as far as
        # each individual absorption dataset is concerned, namely when the
        # wavelengths span multiple datasets we for-loop over them.
        w = np.atleast_1d(w)
        if self.has_absorption:
            values = eval_sigma_a_mono_impl(  # values at altitude levels
                self.absorption_data,
                thermoprops=self._thermoprops_interp(zgrid),
                w=w,
                error_handler_config=self.error_handler_config,
            )  # axis order = (w, z)
            # project on altitude layers
            return 0.5 * (values[:, 1:] + values[:, :-1]).squeeze()
        else:
            return np.zeros((w.size, zgrid.n_layers)).squeeze() / ureg.km

    def eval_sigma_a_ckd(
        self,
        w: pint.Quantity,
        g: float,
        zgrid: ZGrid,
    ) -> pint.Quantity:
        # NOTE: this method accepts 'w'-arrays and is vectorized as far as
        # each individual absorption dataset is concerned, namely when the
        # wavelengths span multiple datasets we for-loop over them.
        w = np.atleast_1d(w)
        if self.has_absorption:
            values = eval_sigma_a_ckd_impl(  # values at altitude levels
                self.absorption_data,
                thermoprops=self._thermoprops_interp(zgrid),
                w=w,
                g=g,
                error_handler_config=self.error_handler_config,
            )  # axis order = (w, z)
            # project on altitude layers
            return 0.5 * (values[:, 1:] + values[:, :-1]).squeeze()
        else:
            return np.zeros((w.size, zgrid.n_layers)).squeeze() / ureg.km

    def eval_sigma_s_mono(self, w: pint.Quantity, zgrid: ZGrid) -> pint.Quantity:
        if self.has_scattering:
            thermoprops = self._thermoprops_interp(zgrid)
            sigma_s = compute_sigma_s_air(
                wavelength=w,
                number_density=to_quantity(thermoprops.n),
            )
            # project on altitude layers
            return 0.5 * (sigma_s[1:] + sigma_s[:-1])
        else:
            return np.zeros((1, zgrid.n_layers)) / ureg.km

    def eval_sigma_s_ckd(
        self,
        w: pint.Quantity,
        g: float,
        zgrid: ZGrid,
    ) -> pint.Quantity:
        return self.eval_sigma_s_mono(w=w, zgrid=zgrid)

    def eval_sigma_t_mono(self, w: pint.Quantity, zgrid: ZGrid) -> pint.Quantity:
        sigma_a = self.eval_sigma_a_mono(w=w, zgrid=zgrid)
        sigma_s = self.eval_sigma_s_mono(w=w, zgrid=zgrid)
        return sigma_a + sigma_s

    def eval_sigma_t_ckd(
        self,
        w: pint.Quantity,
        g: float,
        zgrid: ZGrid,
    ) -> pint.Quantity:
        sigma_a = self.eval_sigma_a_ckd(w=w, g=g, zgrid=zgrid)
        sigma_s = self.eval_sigma_s_ckd(w=w, g=g, zgrid=zgrid)
        return sigma_a + sigma_s

    def eval_dataset_mono(self, w: pint.Quantity, zgrid: ZGrid) -> xr.Dataset:
        return make_dataset(
            wavelength=w,
            z_level=zgrid.levels,
            z_layer=zgrid.layers,
            sigma_a=self.eval_sigma_a_mono(w, zgrid),
            sigma_s=self.eval_sigma_s_mono(w, zgrid),
        ).squeeze()

    def eval_dataset_ckd(
        self,
        w: pint.Quantity,
        g: float,
        zgrid: ZGrid,
    ) -> xr.Dataset:
        return make_dataset(
            wavelength=w,
            z_level=zgrid.levels,
            z_layer=zgrid.layers,
            sigma_a=self.eval_sigma_a_ckd(w=w, g=g, zgrid=zgrid),
            sigma_s=self.eval_sigma_s_ckd(w=w, g=g, zgrid=zgrid),
        ).squeeze()



import numpy as np
import pytest

import eradiate
from eradiate import unit_registry as ureg

# ------------------------------------------------------------------------------
#                            Pre-process helpers
# ------------------------------------------------------------------------------

pytest.register_assert_rewrite("eradiate.test_tools.types.check_scene_element")

# ------------------------------------------------------------------------------
#                                 Mode fixtures
# ------------------------------------------------------------------------------


def generate_fixture(mode):
    @pytest.fixture()
    def fixture():
        import eradiate

        eradiate.set_mode(mode)

    globals()["mode_" + mode] = fixture


for mode in eradiate.modes():
    generate_fixture(mode)
del generate_fixture


def generate_fixture_group(name, modes):
    @pytest.fixture(params=modes)
    def fixture(request):
        mode = request.param
        import eradiate

        eradiate.set_mode(mode)

    globals()["modes_" + name] = fixture


variants = [x for x in eradiate.modes() if x not in {"mono", "ckd"}]  # Remove aliases
variant_groups = {
    "all_mono": [x for x in variants if x.startswith("mono")],
    "all_ckd": [x for x in variants if x.startswith("ckd")],
    "all_mono_ckd": [
        x for x in variants if (x.startswith("mono") or x.startswith("ckd"))
    ],
    "all_single": [x for x in variants if x.endswith("single")],
    "all_double": [x for x in variants if x.endswith("double")],
    "all": variants,
}

for name, variants in variant_groups.items():
    generate_fixture_group(name, variants)
del generate_fixture_group


# ------------------------------------------------------------------------------
#                              Other configuration
# ------------------------------------------------------------------------------


@pytest.fixture
def ert_seed_state():
    from eradiate.rng import SeedState

    return SeedState(0)


def _error_handler_config():
    return {
        "x": {
            "missing": "ignore",
            "scalar": "ignore",
            "bounds": "raise",
        },
        "p": {"bounds": "ignore"},
        "t": {"bounds": "ignore"},
    }


@pytest.fixture
def error_handler_config():
    """Error handler configuration for absorption coefficient interpolation.

    Notes
    -----
    This configuration is chosen to ignore all interpolation issues (except
    bounds error along the mole fraction dimension) because warnings are
    captured by pytest which will raise.
    Ignoring the bounds on pressure and temperature is safe because
    out-of-bounds values usually correspond to locations in the atmosphere
    that are so high that the contribution to the absorption coefficient
    are negligible at these heights.
    The bounds error for the 'x' (mole fraction) coordinate is considered
    fatal.
    """
    return _error_handler_config()


@pytest.fixture
def us_standard_mono():
    """
    AFGL (1986) U.S. Standard atmosphere with monochromatic absorption data.

    Notes
    -----
    Molecules included are H2O, CO2, O3, N2O, CO, CH4, O2.
    Specified absorption data covers the wavelength range [250, 3125] nm.
    Altitude grid is regular with a 1 km step, from 0 to 120 km.
    """
    return {
        "type": "molecular",
        "thermoprops": {
            "identifier": "afgl_1986-us_standard",
            "z": np.linspace(0.0, 120.0, 121) * ureg.km,
            "additional_molecules": False,
        },
        "absorption_data": ("komodo", [549.5, 550.5] * ureg.nm),
        "error_handler_config": _error_handler_config(),
    }


@pytest.fixture
def us_standard_ckd_550nm():
    """
    AFGL (1986) U.S. Standard atmosphere with CKD absorption data.

    Notes
    -----
    Molecules included are H2O, CO2, O3, N2O, CO, CH4, O2.
    Specified absorption data covers the CKD band associated with wavenumber
    interval [18100, 18200] cm^-1, i.e. the wavelenght range
    [549.45, 552.48] nm.
    Altitude grid is regular with a 1 km step, from 0 to 120 km.
    """
    return {
        "type": "molecular",
        "thermoprops": {
            "identifier": "afgl_1986-us_standard",
            "z": np.linspace(0.0, 120.0, 121) * ureg.km,
            "additional_molecules": False,
        },
        "absorption_data": ("monotropa", [549.5, 550.5] * ureg.nm),
        "error_handler_config": _error_handler_config(),
    }


@pytest.fixture
def cams_lybia4_ckd_550nm():
    """
    CAMS Lybia4 atmosphere with CKD absorption data.

    Notes
    -----
    Molecules included are H2O, CO2, O3, N2O, CO, CH4, O2, NO2, NO, SO2.
    Specified absorption data covers the CKD band associated with wavenumber
    interval [18100, 18200] cm^-1, i.e. the wavelenght range
    [549.45, 552.48] nm.
    Altitude grid is regular with a 1 km step, from 0 to 120 km.
    """
    thermoprops = eradiate.data.load_dataset(
        "tests/thermoprops/cams_lybia4_2005-04-01.nc"
    )

    return {
        "type": "molecular",
        "thermoprops": thermoprops,
        "absorption_data": ("monotropa", [549.5, 550.5] * ureg.nm),
        "error_handler_config": _error_handler_config(),
    }



import numpy as np
import pytest

import eradiate
from eradiate import unit_registry as ureg


@pytest.mark.parametrize(
    "atm_flags",
    [
        {"molecular": {}},
        {"molecular": {"has_scattering": False}},
        {"molecular": {"has_absorption": False}},
    ],
    ids=["default", "mol_nosc", "mol_noab"],
)
def test_heterogeneous_atm_flags(
    modes_all_double,
    atm_flags,
    us_standard_mono,
    us_standard_ckd_550nm,
):
    """
    Heterogeneous atmosphere component flags
    ========================================

    This system test tries various combinations of atmosphere flags and checks
    that a computation succeeds and returns non-zero values.
    """
    # TODO: Recycle this test
    if eradiate.mode().is_mono:
        molecular = us_standard_mono
    elif eradiate.mode().is_ckd:
        molecular = us_standard_ckd_550nm
    else:
        raise NotImplementedError

    molecular.update(atm_flags["molecular"])
    exp = eradiate.experiments.AtmosphereExperiment(
        surface={"type": "rpv"},
        atmosphere={
            "type": "heterogeneous",
            "molecular_atmosphere": molecular,
            "particle_layers": {
                "tau_ref": 0.2,
                "bottom": 0.0,
                "top": 10.0 * ureg.km,
            },
        },
        illumination={
            "type": "directional",
            "zenith": 30.0,
            "azimuth": 0.0,
        },
        measures={
            "type": "mdistant",
            "construct": "hplane",
            "azimuth": 0.0,
            "zeniths": np.arange(-75, 76, 5),
            "srf": eradiate.scenes.spectra.MultiDeltaSpectrum(
                wavelengths=550.0 * ureg.nm
            ),
        },
    )

    assert np.all(eradiate.run(exp) > 0.0)



"""
Convert SVG figures to PNG. Requires Inkscape 1.0 (with new CLI).
"""

import subprocess
from pathlib import Path

FIG_PATH = Path(__file__).parent / "fig"


def svg_to_png():
    svg_files = list(FIG_PATH.glob("*.svg"))
    for file in svg_files:
        cmd = [
            "inkscape",
            "--export-type=png",
            "--export-dpi=96",
            str(file.absolute()),
        ]
        print(" ".join(cmd))
        subprocess.run(cmd)


if __name__ == "__main__":
    svg_to_png()



import mitsuba as mi
import pytest

from eradiate import KernelContext
from eradiate.scenes.core import traverse
from eradiate.scenes.measure import PerspectiveCameraMeasure
from eradiate.test_tools.types import check_scene_element


@pytest.mark.parametrize(
    "tested, expected",
    [
        ({}, None),
        (
            {"origin": [0, 0, 0], "target": [0, 0, 0], "up": [0, 0, 1]},
            ValueError,
        ),
        (
            {"origin": [1, 1, 1], "target": [1, 1, 1], "up": [0, 0, 1]},
            ValueError,
        ),
        (
            {"origin": [-1, 0.5, 1.5], "target": [-1, 0.5, 1.5], "up": [0, 0, 1]},
            ValueError,
        ),
        (
            {"origin": [0, 1, 0], "target": [1, 0, 0], "up": [1, -1, 0]},
            ValueError,
        ),
    ],
    ids=[
        "no_args",
        "same_origin_target_1",  # origin and target cannot be the same
        "same_origin_target_2",
        "same_origin_target_3",
        "same_up_direction",  # up and viewing direction cannot be the same
    ],
)
def test_perspective_construct(mode_mono, tested, expected):
    # Constructor
    if expected is None:
        measure = PerspectiveCameraMeasure(**tested)
        check_scene_element(measure, mi.Sensor)

    elif issubclass(expected, ValueError):
        with pytest.raises(expected):
            PerspectiveCameraMeasure(**tested)

    else:
        RuntimeError("unhandled expected value")


def test_perspective_medium(mode_mono):
    measure = PerspectiveCameraMeasure()
    template, _ = traverse(measure)

    kdict = template.render(ctx=KernelContext())
    assert "medium" not in kdict

    kdict = template.render(
        ctx=KernelContext(kwargs={"measure.atmosphere_medium_id": "test_atmosphere"})
    )
    assert kdict["medium"] == {"type": "ref", "id": "test_atmosphere"}



import os
import sys
from copy import deepcopy

import click
import networkx as nx
from _utils import resolve_package_manager
from ruamel.yaml import YAML
from ruamel.yaml.comments import CommentedSeq as CS


@click.command()
@click.option(
    "-s",
    "--sections",
    default="dependencies,main,recommended,tests,dev,docs,optional",
    help="Dependency sections to include in the produced environment.yml file. "
    "Default: 'dependencies,main,recommended,tests,dev,docs,optional'",
)
@click.option(
    "-o", "--output-dir", default="requirements/conda", help="Path to output file."
)
@click.option(
    "-l",
    "--layered-config",
    default="./requirements/layered.yml",
    help="Path to layered requirement dependency configuration file. "
    "Default: ./requirements/layered.yml",
)
@click.option(
    "-p",
    "--pip-deps",
    is_flag=True,
    default=False,
    help="Include pip dependencies in the environment.yml file. Default: False",
)
@click.option("-q", "--quiet", is_flag=True, help="Suppress terminal output.")
@click.option(
    "-g",
    "--write-graphs",
    is_flag=True,
    default=False,
    help="Write dependency graphs to Graphviz dot files. Default: False",
)
def cli(sections, output_dir, layered_config, pip_deps, quiet, write_graphs):
    """
    Create a Conda environment file from a pyproject.toml.
    """

    # Set YAML parameters
    yaml = YAML(typ="rt")  # Round-trip mode allows for comment insertion
    indent_offset = 2
    yaml.indent(offset=indent_offset)

    # Load environment file template
    with open(os.path.join("requirements", "environment.in")) as f:
        env_yml = yaml.load(f.read())
    with open(layered_config) as f:
        layered_yml = resolve_package_manager(yaml.load(f.read()), "conda")

    sections = [
        (x.strip(), {"packages": set(layered_yml[x.strip()].get("packages", []))})
        for x in sections.split(",")
    ]
    dep_list = deepcopy(env_yml["dependencies"]) if "dependencies" in env_yml else []

    # Create dependency graph
    G = nx.DiGraph()
    G.add_nodes_from(sections)
    for section, data in sections:
        for include in layered_yml[section].get("includes", []):
            G.add_edge(section, include)

    if write_graphs:
        path = os.path.join(output_dir, "layer_graph.dot")
        nx.drawing.nx_pydot.write_dot(G, path)

    G.add_node("default", packages=set(dep_list))
    G.add_edge("main", "default")

    if not nx.algorithms.dag.is_directed_acyclic_graph(G):
        raise ValueError("Dependency graph is not a DAG")

    sections = list(nx.algorithms.dag.topological_sort(G))[::-1]

    for section in sections:
        # Select sections to include
        subG = G.subgraph(nx.algorithms.dag.descendants(G, section) | {section})
        environment_packages = dict(sorted(subG.nodes.data()))

        # Create dependency list
        section_indices = dict()
        added_packages = set()
        ordered_packages = []
        for included_section, data in environment_packages.items():
            packages_to_add = sorted(data["packages"] - added_packages)

            # Temporary fix until an eradiate_mitsuba conda package is made available
            # This is disabled by default because the Poetry solver fails to find the
            # wheel for DrJIT on OSX.
            mitsuba_pkg = set(p for p in packages_to_add if "eradiate-mitsuba" in p)
            if mitsuba_pkg:
                packages_to_add = [p for p in packages_to_add if p not in mitsuba_pkg]

            section_indices[included_section] = len(ordered_packages)
            if packages_to_add:
                ordered_packages.extend(packages_to_add)

            # Temporary fix until an eradiate_mitsuba conda package is made available
            # This is disabled by default because the Poetry solver fails to find the
            # wheel for DrJIT on OSX.
            if mitsuba_pkg and pip_deps:
                if "pip" not in added_packages:
                    ordered_packages.append("pip")
                ordered_packages.append({"pip": list(mitsuba_pkg)})

            added_packages |= data["packages"]

        if section == "default":
            continue

        # Format dependency list
        lst = CS(ordered_packages)
        for included_section, i in section_indices.items():
            lst.yaml_set_comment_before_after_key(i, included_section, indent_offset)
        env_yml["dependencies"] = lst

        # Output to terminal
        if not quiet:
            yaml.dump(env_yml, sys.stdout)

        # Output to file
        if output_dir is not None:
            path = os.path.join(output_dir, f"environment-{section}.yml")
            with open(path, "w") as outfile:
                if not quiet:
                    print()
                print(f"Saving to {path}")
                yaml.dump(env_yml, outfile)

        if write_graphs:
            path = os.path.join(output_dir, f"environment-{section}.dot")
            if "default" in subG:
                subG = subG.copy()
                subG.remove_node("default")
            nx.drawing.nx_pydot.write_dot(subG, path)


if __name__ == "__main__":
    cli()



import joseki
import numpy as np
import xarray as xr

from eradiate import unit_registry as ureg
from eradiate.radprops.rayleigh import (
    _LOSCHMIDT,
    air_refractive_index,
    compute_sigma_s_air,
)
from eradiate.units import to_quantity


def test_sigma_s_air():
    """
    Test computation of Rayleigh scattering coefficient for air with
    default values.
    """

    ref_cross_section = ureg.Quantity(4.513e-27, "cm**2")
    ref_sigmas = ref_cross_section * _LOSCHMIDT
    expected = ref_sigmas

    # Compare to reference value computed from scattering cross section in
    # Bates (1984) Planetary and Space Science, Volume 32, No. 6.
    assert np.allclose(
        compute_sigma_s_air(number_density=_LOSCHMIDT), expected, rtol=1e-2
    )


def test_sigma_s_air_wavelength_dependence():
    """
    Test that the Rayleigh scattering coefficient scales with the 4th power
    of wavelength.
    """

    wavelength = ureg.Quantity(np.linspace(240.0, 2400.0), "nm")
    sigma_s = compute_sigma_s_air(wavelength)
    prod = sigma_s.magnitude * np.power(wavelength, 4)
    assert np.allclose(prod, prod[0], rtol=0.25)


def test_sigma_s_air_optical_thickness():
    """
    We compute the total optical thickness due to Rayleigh scattering by the
    air in a 100km high atmosphere. We compare the obtained result to the value
    given by :cite:`Hansen1974LightScatteringPlanetary` at page 544.
    """
    thermoprops = joseki.make(
        identifier="ussa_1976",
        z=np.linspace(0.0, 100000.0, 1001) * ureg.m,
    )
    n = to_quantity(thermoprops.n)
    z = to_quantity(thermoprops.z)
    sigma_s = compute_sigma_s_air(number_density=n)

    sigma_s = xr.DataArray(
        data=sigma_s.m_as("km^-1"),
        dims=("z",),
        coords={
            "z": ("z", z.m_as("km")),
        },
        attrs=dict(units=sigma_s.units),
    )
    optical_thickness = sigma_s.integrate(coord="z")

    assert np.isclose(optical_thickness, 0.0973, rtol=1e-2)


def test_air_refractive_index():
    """
    Test computation of the air refractive index for different wavelength
    values. We compare the results with the values given in Table III of
    :cite:`Peck1972DispersionAir`.
    """

    wavelength = ureg.Quantity(
        np.array([1.6945208, 1.01425728, 0.64402492, 0.54622707, 0.3889751, 0.230289]),
        "micrometer",
    )

    indices = air_refractive_index(wavelength)

    # compute the refractivities in parts per 1e8
    results = (indices - 1) * 1e8

    expected = np.array([27314.19, 27410.90, 27638.092, 27789.843, 28336.843, 30787.68])
    assert np.allclose(results, expected, rtol=1e-5)


def test_air_refractive_index_multidim():
    """
    Supports multiple wavelength and number density values.
    """
    w = ureg.Quantity(np.linspace(280, 2400), "nm")
    n = ureg.Quantity(np.array([_LOSCHMIDT.m_as("m^-3")] * 8), "m^-3")
    result = air_refractive_index(wavelength=w, number_density=n)
    assert result.shape == (len(w), len(n))


def test_compute_sigma_s_air_multidim():
    """
    Supports multiple wavelength and number density values.
    """
    w = ureg.Quantity(np.linspace(280, 2400), "nm")
    n = ureg.Quantity(np.array([_LOSCHMIDT.m_as("m^-3")] * 8), "m^-3")
    result = compute_sigma_s_air(wavelength=w, number_density=n)
    assert result.shape == (len(w), len(n))



from __future__ import annotations

from collections import OrderedDict
from pathlib import Path

from ._blind_directory import BlindDirectoryDataStore
from ._blind_online import BlindOnlineDataStore
from ._multi import MultiDataStore
from ._safe_directory import SafeDirectoryDataStore
from ._safe_online import SafeOnlineDataStore
from .._config import config

#: Global data store.
data_store: MultiDataStore | None = None


def init_data_store(
    offline: bool | None = None, production: bool | None = None
) -> None:
    """
    Initialize the global data store.

    Parameters
    ----------
    offline : bool, optional
        If ``True``, replace all online data stores with blind directory data
        stores. If unset, the global offline configuration is used.

    production : bool, optional
        If ``True``, replace all development data stores providing files from
        the current local source folder with online safe data stores. If set to
        ``True`` with ``offline`` flag set to ``True``, then the data must be
        provided by the user before accessing it. If unset, the global
        production configuration is used.

    Notes
    -----
    This function is called automatically when the ``eradiate.data`` package is
    imported.
    """
    global data_store

    if offline is None:
        offline = config.offline

    if production is None:
        production = config.source_dir is None

    download_dir = config.download_dir
    if download_dir is None:
        if production:
            download_dir = Path("./eradiate_downloads").absolute()
        else:
            download_dir = config.source_dir / "resources" / "data"

    if not production:
        small_files = SafeDirectoryDataStore(path=download_dir)
    else:
        if offline:
            small_files = BlindDirectoryDataStore(path=download_dir)
        else:
            small_files = SafeOnlineDataStore(
                base_url="/".join(
                    [
                        config.small_files_registry_url,
                        config.small_files_registry_revision,
                    ]
                ),
                path=download_dir,
            )

    if offline:
        data_store = MultiDataStore(
            stores=OrderedDict(
                [
                    ("small_files", small_files),
                    (
                        "large_files_stable",
                        BlindDirectoryDataStore(path=download_dir / "stable"),
                    ),
                    (
                        "large_files_unstable",
                        BlindDirectoryDataStore(path=download_dir / "unstable"),
                    ),
                ]
            )
        )

    else:
        data_store = MultiDataStore(
            stores=OrderedDict(
                [
                    ("small_files", small_files),
                    (
                        "large_files_stable",
                        SafeOnlineDataStore(
                            base_url="/".join([config.data_store_url, "stable"]),
                            path=download_dir / "stable",
                        ),
                    ),
                    (
                        "large_files_unstable",
                        BlindOnlineDataStore(
                            base_url="/".join([config.data_store_url, "unstable"]),
                            path=download_dir / "unstable",
                        ),
                    ),
                ]
            )
        )


# Initialise the data store upon module import
init_data_store()



import os
import tempfile

import mitsuba as mi
import pytest

from eradiate import KernelContext
from eradiate import unit_registry as ureg
from eradiate.kernel._kernel_dict import KernelDictTemplate
from eradiate.scenes.biosphere import MeshTree, MeshTreeElement
from eradiate.scenes.core import traverse
from eradiate.test_tools.types import check_scene_element
from eradiate.util.misc import flatten

# ------------------------------------------------------------------------------
#                            Fixture definitions
# ------------------------------------------------------------------------------


@pytest.fixture(scope="module")
def tmpfile_obj():
    with tempfile.TemporaryDirectory() as tmpdir:
        filename = os.path.join(tmpdir, "tmpfile_mesh.obj")
        with open(filename, "w") as tf:
            tf.write(
                """o Cube
v 1.000000 -1.000000 -1.000000
v 1.000000 -1.000000 1.000000
v -1.000000 -1.000000 1.000000
v -1.000000 -1.000000 -1.000000
v 1.000000 1.000000 -0.999999
v 0.999999 1.000000 1.000001
v -1.000000 1.000000 1.000000
v -1.000000 1.000000 -1.000000
vt 1.000000 0.333333
vt 1.000000 0.666667
vt 0.666667 0.666667
vt 0.666667 0.333333
vt 0.666667 0.000000
vt 0.000000 0.333333
vt 0.000000 0.000000
vt 0.333333 0.000000
vt 0.333333 1.000000
vt 0.000000 1.000000
vt 0.000000 0.666667
vt 0.333333 0.333333
vt 0.333333 0.666667
vt 1.000000 0.000000
vn 0.000000 -1.000000 0.000000
vn 0.000000 1.000000 0.000000
vn 1.000000 0.000000 0.000000
vn -0.000000 0.000000 1.000000
vn -1.000000 -0.000000 -0.000000
vn 0.000000 0.000000 -1.000000
s off
f 2/1/1 3/2/1 4/3/1
f 8/1/2 7/4/2 6/5/2
f 5/6/3 6/7/3 2/8/3
f 6/8/4 7/5/4 3/4/4
f 3/9/5 7/10/5 8/11/5
f 1/12/6 4/13/6 8/11/6
f 1/4/1 2/1/1 4/3/1
f 5/14/2 8/1/2 6/5/2
f 1/12/3 5/6/3 2/8/3
f 2/12/4 6/8/4 3/4/4
f 4/13/5 3/9/5 8/11/5
f 5/6/6 1/12/6 8/11/6"""
            )
        yield filename


@pytest.fixture(scope="module")
def tmpfile_ply():
    with tempfile.TemporaryDirectory() as tmpdir:
        filename = os.path.join(tmpdir, "tmpfile_mesh.ply")
        with open(filename, "w") as tf:
            tf.write(
                """ply
format ascii 1.0
element vertex 8
property float x
property float y
property float z
element face 12
property list uchar int32 vertex_index
end_header
0 0 0
0 0 1
0 1 1
0 1 0
1 0 0
1 0 1
1 1 1
1 1 0
3 0 1 2
3 0 2 3
3 7 6 5
3 7 5 4
3 0 4 5
3 0 5 1
3 1 5 6
3 1 6 2
3 2 6 7
3 2 7 3
3 3 7 4
3 3 4 0
"""
            )
        yield filename


@pytest.fixture(scope="module")
def tmpfile_stl():
    with tempfile.TemporaryDirectory() as tmpdir:
        filename = os.path.join(tmpdir, "tmpfile_mesh.stl")
        with open(filename, "w") as tf:
            tf.write("fake content\n")
        yield filename


# ------------------------------------------------------------------------------
#                                   Tests
# ------------------------------------------------------------------------------


@pytest.mark.parametrize(
    "tmp_file",
    ["tmpfile_obj", "tmpfile_ply", "tmpfile_stl"],
    ids=["obj", "ply", "stl"],
)
def test_mesh_tree_element_instantiate(mode_mono, tmp_file, request):
    tmp_file = request.getfixturevalue(tmp_file)

    # empty constructor raises due to missing mesh description file
    with pytest.raises(TypeError):
        MeshTreeElement()

    # instantiation with supported format succeeds, except for stl
    if "tmpfile_stl" not in request.fixturenames:
        assert MeshTreeElement(
            id="mesh_tree",
            mesh_filename=tmp_file,
            mesh_units=ureg.m,
            reflectance=0.5,
            transmittance=0.5,
        )
    else:
        with pytest.raises(ValueError):
            MeshTreeElement(
                id="mesh_tree_stl",
                mesh_filename=tmp_file,
                mesh_units=ureg.m,
                reflectance=0.5,
                transmittance=0.5,
            )


@pytest.mark.parametrize("tmp_file", ["tmpfile_obj", "tmpfile_ply"], ids=["obj", "ply"])
def test_mesh_tree_element_load(mode_mono, tmp_file, request):
    """
    Instantiate MeshTreeElement objects from obj and ply files and load the
    corresponding Mitsuba objects.
    """
    tmp_file = request.getfixturevalue(tmp_file)

    tree_element = MeshTreeElement(
        id="mesh_tree_obj",
        mesh_filename=tmp_file,
        mesh_units=ureg.m,
        reflectance=0.5,
        transmittance=0.5,
    )
    template = KernelDictTemplate(
        flatten(
            {
                "type": "scene",
                **tree_element._template_bsdfs,
                **tree_element._template_shapes,
            }
        )
    )

    # Check that the template successfully expands to a valid Mitsuba scene
    mi_scene = mi.load_dict(template.render(KernelContext()))
    assert isinstance(mi_scene, mi.Scene)


def test_mesh_tree_instantiate(mode_mono, tmpfile_obj):
    """
    Instantiate a MeshTree object holding two MeshTreeElements and load the
    corresponding Mitsuba objects.
    """

    # Constructor based instantiation
    assert MeshTree(
        mesh_tree_elements=[
            MeshTreeElement(
                id="foliage",
                mesh_filename=tmpfile_obj,
                mesh_units=ureg.m,
                reflectance=0.5,
                transmittance=0.5,
            ),
            MeshTreeElement(
                id="trunk",
                mesh_filename=tmpfile_obj,
                mesh_units=ureg.m,
                reflectance=0.9,
                transmittance=0.1,
            ),
        ]
    )

    tree = MeshTree(
        mesh_tree_elements=[
            MeshTreeElement(
                id="foliage",
                mesh_filename=tmpfile_obj,
                mesh_units=ureg.m,
                reflectance=0.5,
                transmittance=0.5,
            ),
            MeshTreeElement(
                id="trunk",
                mesh_filename=tmpfile_obj,
                mesh_units=ureg.m,
                reflectance=0.9,
                transmittance=0.1,
            ),
        ]
    )

    check_scene_element(tree)
    _, params = traverse(tree)
    assert set(params.keys()) == {
        "bsdf_foliage.reflectance.value",
        "bsdf_foliage.transmittance.value",
        "bsdf_trunk.reflectance.value",
        "bsdf_trunk.transmittance.value",
    }

    # check that special converter for the mesh_tree_elements field works as
    # expected
    tree = MeshTree(
        id="mesh_tree",
        mesh_tree_elements=[
            {
                "id": "tree_foliage",
                "mesh_filename": tmpfile_obj,
                "mesh_units": "m",
                "reflectance": 0.5,
                "transmittance": 0.5,
            },
            {
                "id": "tree_trunk",
                "mesh_filename": tmpfile_obj,
                "mesh_units": "m",
                "reflectance": 0.9,
                "transmittance": 0.1,
            },
        ],
    )

    check_scene_element(tree)
    _, params = traverse(tree)
    assert set(params.keys()) == {
        "bsdf_tree_foliage.reflectance.value",
        "bsdf_tree_foliage.transmittance.value",
        "bsdf_tree_trunk.reflectance.value",
        "bsdf_tree_trunk.transmittance.value",
    }



import mitsuba as mi

from eradiate import KernelContext
from eradiate.scenes.core import traverse
from eradiate.scenes.measure import RadiancemeterMeasure
from eradiate.test_tools.types import check_scene_element


def test_radiancemeter_construct(mode_mono):
    measure = RadiancemeterMeasure()
    check_scene_element(measure, mi.Sensor)


def test_radiancemeter_medium(mode_mono):
    measure = RadiancemeterMeasure()
    template, _ = traverse(measure)

    ctx1 = KernelContext()
    ctx2 = KernelContext(kwargs={"measure.atmosphere_medium_id": "test_atmosphere"})

    kdict = template.render(ctx=ctx1)
    assert "medium" not in kdict

    kdict = template.render(ctx=ctx2)
    assert kdict["medium"] == {"type": "ref", "id": "test_atmosphere"}



import typing as t

import pandas as pd
import xarray as xr

XarrayObj = t.Union[xr.DataArray, xr.Dataset]


def unstack_mdistant_grid(obj: XarrayObj) -> XarrayObj:
    """
    Reindex data produced by a :meth:`.MultiDistantMeasure.grid()` measure on
    (VAA, VZA) dimensions.

    This effectively reshapes the data, which is initially indexed on film width
    and height dimensions (the latter having a single element).

    Parameters
    ----------
    obj : DataArray or Dataset
        An xarray object produced by processing a
        :meth:`.MultiDistantMeasure.grid()` measure, indexed on film dimensions.

    Returns
    -------
    DataArray or Dataset
        An object of the same type, reindexed on VZA and VAA coordinates.
    """
    # Collect metadata
    vza_attrs = obj["vza"].attrs
    vaa_attrs = obj["vaa"].attrs

    # Build new index
    idx = pd.MultiIndex.from_arrays(
        (obj.vza.values.flatten(), obj.vaa.values.flatten()), names=("vza", "vaa")
    )

    # Reindex object
    result = (
        obj.drop_vars(
            # First remove the target coords, as well as the indexing dim and
            # associated coords
            # Also remove the x film coordinate (irrelevant after unstacking)
            ("vza", "vaa", "x_index", "x")
        )
        .reindex(x_index=idx)
        .unstack()
    )  # Apply the new index and unstack

    # Reapply metadata
    result["vza"].attrs = vza_attrs
    result["vaa"].attrs = vaa_attrs

    return result



"""Test cases targeting the phase function component of a 1D experiment."""
import os

import matplotlib.pyplot as plt
import numpy as np
import pytest

import eradiate
from eradiate import unit_registry as ureg
from eradiate.experiments import AtmosphereExperiment
from eradiate.units import to_quantity


def make_figure(fname_plot, brf_1, brf_2):
    fig = plt.figure(figsize=(8, 5))
    params = dict(x="vza", ls="dotted", marker=".")
    brf_1.plot(**params)
    brf_2.plot(**params)
    plt.legend(["experiment 1 (rayleigh)", "experiment 2 (tabphase)"])
    plt.title(f"w = {to_quantity(brf_1.w).squeeze():~}")
    plt.tight_layout()

    print(f"Saving figure to {fname_plot}")
    fig.savefig(fname_plot, dpi=200)
    plt.close()


@pytest.mark.slow
def test(
    mode_mono_double,
    rayleigh_tab_phase,
    artefact_dir,
    ert_seed_state,
    request,
):
    r"""
    Equivalency of plugin and tabulated versions of Rayleigh phase function
    =======================================================================

    Assert that a homogeneous atmosphere with a Rayleigh phase function is
    equivalent to a homogeneous atmosphere with a tabulated version of the
    Rayleigh phase function.

    Rationale
    ---------

    * Sensor: Distant measure covering a plane (11 angular points,
      10^6 sample per pixel) and targeting (0, 0, 0).
    * Illumination: Directional illumination with a zenith angle
      :math:`\theta = 30.0°` with black body irradiance spectrum.
    * Surface: a square surface with a Lambertian BRDF with 35 % reflectance.
    * Atmosphere: a non-absorbing 10 km high homogeneous atmosphere. The
      scattering phase function is defined in two ways in two experiments:
      * in experiment 1: using
        :class:`~eradiate.scenes.phase.RayleighPhaseFunction`
      * in experiment 2: using
        :class:`~eradiate.scenes.phase.TabulatedPhaseFunction` with Rayleigh
        phase function tabulated values.
    * Integrator: volumetric path tracer.

    Expected behaviour
    ------------------

    Distant bidirectional reflectance factors measured by both experiments must
    agree within 0.5%.

    Results
    -------

    .. image:: generated/plots/test_onedim_phase_300.0.png
       :width: 95%
    .. image:: generated/plots/test_onedim_phase_550.0.png
       :width: 95%
    .. image:: generated/plots/test_onedim_phase_650.0.png
       :width: 95%
    .. image:: generated/plots/test_onedim_phase_850.0.png
       :width: 95%
    .. image:: generated/plots/test_onedim_phase_1000.0.png
       :width: 95%
    .. image:: generated/plots/test_onedim_phase_1500.0.png
       :width: 95%
    .. image:: generated/plots/test_onedim_phase_2000.0.png
       :width: 95%
    .. image:: generated/plots/test_onedim_phase_2400.0.png
       :width: 95%
    """
    wavelengths = [280, 400, 550, 650, 1000, 1500, 2400] * ureg.nm
    spp = int(1e6)
    reflectance = 0.35
    bottom = 0.0 * ureg.km
    top = 10.0 * ureg.km

    def init_experiment(sigma_a, sigma_s, phase):
        assert eradiate.mode().id == "mono_double"

        return AtmosphereExperiment(
            geometry={
                "type": "plane_parallel",
                "ground_altitude": bottom,
                "toa_altitude": top,
            },
            measures={
                "type": "mdistant",
                "construct": "hplane",
                "srf": {"type": "multi_delta", "wavelengths": wavelengths},
                "zeniths": np.linspace(-75, 75, 11) * ureg.deg,
                "azimuth": 0.0 * ureg.deg,
            },
            illumination={
                "type": "directional",
                "zenith": 30 * ureg.deg,
                "azimuth": 0.0 * ureg.deg,
                "irradiance": {"type": "solar_irradiance", "dataset": "blackbody_sun"},
            },
            atmosphere={
                "type": "homogeneous",
                "sigma_a": sigma_a,
                "sigma_s": sigma_s,
                "phase": phase,
            },
            surface={"type": "lambertian", "reflectance": reflectance},
        )

    experiment_1 = init_experiment(
        sigma_a=0.0,
        sigma_s=eradiate.scenes.spectra.AirScatteringCoefficientSpectrum(),
        phase=eradiate.scenes.phase.RayleighPhaseFunction(),
    )

    ert_seed_state.reset()
    eradiate.run(experiment_1, seed_state=ert_seed_state, spp=spp)

    radprops = rayleigh_tab_phase(albedo=1.0)
    w = to_quantity(radprops.w)
    sigma_t = to_quantity(radprops.sigma_t)
    albedo = to_quantity(radprops.albedo)
    sigma_s_2 = sigma_t * albedo
    sigma_a_2 = sigma_t * (1.0 - albedo)
    phase_2 = eradiate.scenes.phase.TabulatedPhaseFunction(data=radprops.phase)

    experiment_2 = init_experiment(
        sigma_a={
            "type": "interpolated",
            "values": sigma_a_2,
            "wavelengths": w,
        },
        sigma_s={
            "type": "interpolated",
            "values": sigma_s_2,
            "wavelengths": w,
        },
        phase=phase_2,
    )

    ert_seed_state.reset()
    eradiate.run(experiment_2, seed_state=ert_seed_state, spp=spp)

    brf_1 = experiment_1.results["measure"]["brf"]
    brf_2 = experiment_2.results["measure"]["brf"]

    # Make figure
    for w in wavelengths.magnitude:
        filename = f"{request.node.originalname}_{w}.png"
        outdir = os.path.join(artefact_dir, "plots")
        os.makedirs(outdir, exist_ok=True)
        fname_plot = os.path.join(outdir, filename)
        make_figure(fname_plot=fname_plot, brf_1=brf_1.sel(w=w), brf_2=brf_2.sel(w=w))

    np.testing.assert_allclose(brf_1.values, brf_2.values, rtol=5e-3)



import numpy as np
import pytest
import xarray as xr

import eradiate
from eradiate import unit_registry as ureg
from eradiate.units import symbol

eradiate.kernel.install_logging()
eradiate.config.progress = "spectral_loop"


@pytest.fixture(scope="module")
def rayleigh_tab_phase():
    """
    Radiative properties of a Rayleigh scattering medium in [280, 2400] nm.

    The Rayleigh scattering phase function is computed from the analytical
    formula and tabulated on a regular scattering angle cosine grid.
    The Rayleigh scattering phase function does not depend on wavelength.

    The Rayleigh scattering coefficient is computed with
    :meth:`~eradiate.radprops.rayleigh.compute_sigma_s_air` with default
    the number density, i.e., standard air number density.

    The single-scattering albedo can be set but it is assumed to be spectrally
    invariant.

    This fixture returns a callable that takes the single-scattering albedo as
    argument and returns the corresponding Rayleigh radiative properties data
    set.
    """

    def radprops(albedo: float = 1.0, sigma_t=None) -> xr.Dataset:
        w = np.linspace(279.0, 2401.0, 10000) * ureg.nm

        # Collision coefficients
        albedo = albedo * np.ones_like(w) * ureg.dimensionless
        if sigma_t is not None:
            sigma_t = sigma_t.reshape(w.shape)
        else:
            sigma_s = eradiate.radprops.rayleigh.compute_sigma_s_air(wavelength=w)
            sigma_t = (
                np.divide(
                    sigma_s.m,
                    albedo.m,
                    where=albedo.m != 0,
                    out=np.zeros_like(sigma_s),
                )
                * sigma_s.units
            )

        # Phase function
        # Note: rayleigh phase function does not change with wavelength
        mu = np.linspace(-1.0, 1.0)
        arrays = [3.0 * (1 + np.square(mu)) / (16 * np.pi) / ureg.steradian for _ in w]
        phase = np.stack(arrays, axis=0).reshape(w.size, mu.size, 1, 1)

        return xr.Dataset(
            data_vars={
                "sigma_t": (
                    "w",
                    sigma_t.magnitude,
                    {
                        "standard_name": "air_volume_extinction_coefficient",
                        "long_name": "extinction coefficient",
                        "units": symbol(sigma_t.units),
                    },
                ),
                "albedo": (
                    "w",
                    albedo.magnitude,
                    {
                        "standard_name": "single_scattering_albedo",
                        "long_name": "albedo",
                        "units": symbol(albedo.units),
                    },
                ),
                "phase": (
                    ("w", "mu", "i", "j"),
                    phase.magnitude,
                    {
                        "standard_name": "scattering_phase_matrix",
                        "long_name": "phase matrix",
                        "units": symbol(phase.units),
                    },
                ),
            },
            coords={
                "w": (
                    "w",
                    w.magnitude,
                    {
                        "standard_name": "radiation_wavelength",
                        "long_name": "wavelength",
                        "units": symbol(w.units),
                    },
                ),
                "mu": (
                    "mu",
                    mu,
                    {
                        "standard_name": "scattering_angle_cosine",
                        "long_name": "scattering angle cosine",
                        "units": "dimensionless",
                    },
                ),
                "i": ("i", [0]),
                "j": ("j", [0]),
            },
        )

    return radprops



import os
import tempfile

import numpy as np
import pytest

from eradiate import KernelContext
from eradiate import unit_registry as ureg
from eradiate.scenes.biosphere import AbstractTree, LeafCloud
from eradiate.scenes.core import traverse
from eradiate.test_tools.types import check_scene_element

# ------------------------------------------------------------------------------
#                            Fixture definitions
# ------------------------------------------------------------------------------


@pytest.fixture(scope="module")
def tempfile_leaves():
    with tempfile.TemporaryDirectory() as tmpdir:
        filename = os.path.join(tmpdir, "tempfile_leaves.txt")
        with open(filename, "w") as tf:
            tf.write("0.100 8.864 9.040 1.878 -0.314 0.025 0.949\n")
            tf.write("0.100 9.539 -10.463 0.627 0.489 -0.276 0.828\n")
            tf.write("0.100 -2.274 -9.204 0.797 0.618 0.184 0.764\n")
            tf.write("0.100 -9.957 -4.971 0.719 -0.066 0.100 0.993\n")
            tf.write("0.100 5.339 9.153 0.500 0.073 -0.294 0.953\n")
        yield filename


# ------------------------------------------------------------------------------
#                              AbstractTree Tests
# ------------------------------------------------------------------------------


def test_abstract_tree_instantiate(mode_mono):
    """Unit tests for :class:`AbstractTree`'s default constructor."""

    # Empty constructor does not raise
    assert AbstractTree()

    # Now with more sensible values
    assert AbstractTree(
        leaf_cloud=LeafCloud(
            leaf_positions=[[0, 0, 0]], leaf_orientations=[[0, 0, 1]], leaf_radii=[0.1]
        ),
        trunk_height=0.5,
        trunk_radius=0.1,
        trunk_reflectance=0.5,
    )


def test_abstract_tree_dispatch_leaf_cloud(mode_mono, tempfile_leaves):
    """Test if contained LeafCloud is instantiated in all variants"""

    # A LeafCloud instance can be loaded from a file on the hard drive
    tree = AbstractTree(leaf_cloud=LeafCloud.from_file(tempfile_leaves))
    assert len(tree.leaf_cloud.leaf_positions) == 5
    assert np.allclose(tree.leaf_cloud.leaf_radii, 0.1 * ureg.m)
    # Produced kernel dict is valid
    check_scene_element(tree)
    # Check parameter map contents
    _, umap_template = traverse(tree)
    assert set(umap_template.keys()) == {
        "bsdf_leaf_cloud.reflectance.value",
        "bsdf_leaf_cloud.transmittance.value",
        "bsdf_abstract_tree.reflectance.value",
    }

    # When passing a dict for the leaf_cloud field, the 'type' param can be omitted
    assert AbstractTree(
        leaf_cloud={
            "leaf_positions": [[0, 0, 0], [1, 1, 1]],
            "leaf_orientations": [[0, 0, 1], [1, 0, 0]],
            "leaf_radii": [0.1, 0.1],
        }
    )

    # Dispatch to from_file if requested
    tree1 = AbstractTree(leaf_cloud=LeafCloud.from_file(tempfile_leaves))
    tree2 = AbstractTree(
        leaf_cloud={"construct": "from_file", "filename": tempfile_leaves}
    )
    # assert leaf clouds are equal
    assert np.all(tree1.leaf_cloud.leaf_positions == tree2.leaf_cloud.leaf_positions)
    assert np.all(
        tree1.leaf_cloud.leaf_orientations == tree2.leaf_cloud.leaf_orientations
    )
    assert np.all(tree1.leaf_cloud.leaf_radii == tree2.leaf_cloud.leaf_radii)
    assert (
        tree1.leaf_cloud.leaf_transmittance.value
        == tree2.leaf_cloud.leaf_transmittance.value
    )
    assert (
        tree1.leaf_cloud.leaf_reflectance.value
        == tree2.leaf_cloud.leaf_reflectance.value
    )

    # Dispatch to generator if requested
    tree = AbstractTree(
        leaf_cloud={
            "construct": "cuboid",
            "n_leaves": 100,
            "l_horizontal": 10.0,
            "l_vertical": 1.0,
            "leaf_radius": 10.0,
            "leaf_radius_units": "cm",
        }
    )
    assert len(tree.leaf_cloud.leaf_radii) == 100
    assert len(tree.leaf_cloud.leaf_positions) == 100
    assert np.allclose(tree.leaf_cloud.leaf_radii, 10.0 * ureg.cm)


def test_abstract_tree_kernel_dict(mode_mono):
    """Partial unit testing for :meth:`LeafCloud.kernel_dict`."""
    tree_id = "my_tree"
    cloud_id = "my_cloud"
    tree = AbstractTree(
        leaf_cloud=LeafCloud(
            id=cloud_id,
            leaf_positions=[[0, 0, 0], [1, 1, 1]],
            leaf_orientations=[[1, 0, 0], [0, 1, 0]],
            leaf_radii=[0.1, 0.1],
            leaf_reflectance=0.5,
            leaf_transmittance=0.5,
        ),
        id=tree_id,
        trunk_height=2.0,
        trunk_radius=0.2,
        trunk_reflectance=0.5,
    )

    template = traverse(tree)[0]
    kdict = template.render(ctx=KernelContext())

    # The BSDF is bilambertian with the parameters we initially set
    assert kdict[f"bsdf_{cloud_id}"] == {
        "type": "bilambertian",
        "reflectance": {"type": "uniform", "value": 0.5},
        "transmittance": {"type": "uniform", "value": 0.5},
    }

    # Leaves are disks
    for shape_key in [f"{cloud_id}_leaf_0", f"{cloud_id}_leaf_1"]:
        assert kdict[shape_key]["type"] == "disk"

    # Kernel dict is valid
    check_scene_element(tree)



import mitsuba as mi
import pytest

from eradiate.exceptions import TraversalError
from eradiate.scenes.core import Scene, traverse
from eradiate.scenes.shapes import BufferMeshShape
from eradiate.test_tools.types import check_scene_element


@pytest.mark.parametrize(
    "kwargs, expected",
    [
        ({"vertices": [[1, 0, 0], [0, 1, 0], [0, 0, 1]], "faces": [[1, 2, 3]]}, None),
        ({"vertices": [1, 0, 0, 0, 1, 0, 0, 0, 1], "faces": [[1, 2, 3]]}, ValueError),
        (
            {"vertices": [[1, 0, 0], [0, 1, 0], [0, 0, 1]], "faces": [1, 2, 3]},
            ValueError,
        ),
    ],
    ids=["correct", "flat_vertex", "flat_face"],
)
def test_buffer_mesh_construct(modes_all_double, kwargs, expected):
    if expected is None:
        assert BufferMeshShape(**kwargs)

    elif issubclass(expected, Exception):
        with pytest.raises(expected):
            BufferMeshShape(**kwargs)


def test_buffer_mesh_instance(mode_mono):
    mesh = BufferMeshShape(
        vertices=[[1, 0, 0], [0, 1, 0], [0, 0, 1]],
        faces=[[1, 2, 3]],
        bsdf={"type": "lambertian"},
    )
    assert isinstance(mesh.instance, mi.Mesh)


def test_buffer_mesh_params(mode_mono):
    mesh = BufferMeshShape(
        vertices=[[1, 0, 0], [0, 1, 0], [0, 0, 1]],
        faces=[[1, 2, 3]],
        bsdf={"type": "lambertian"},
    )
    assert set(mesh.params.keys()) == {"bsdf.reflectance.value"}


def test_buffer_mesh_traverse(mode_mono):
    mesh = BufferMeshShape(
        vertices=[[1, 0, 0], [0, 1, 0], [0, 0, 1]],
        faces=[[1, 2, 3]],
        bsdf={"type": "lambertian"},
    )

    # Traversal should fail because instance nodes can't be traversed on their own
    with pytest.raises(
        TraversalError, match="Instances may only be inserted as child nodes"
    ):
        traverse(mesh)

    # Traversal as part of an enclosing object should succeed
    scene = Scene(objects={"mesh": mesh})
    check_scene_element(scene, mi.Scene)



import mitsuba as mi
import numpy as np
import pytest

from eradiate import unit_context_config as ucc
from eradiate import unit_registry as ureg
from eradiate.scenes.shapes import SphereShape
from eradiate.test_tools.types import check_scene_element


@pytest.mark.parametrize(
    "kwargs, expected_reflectance",
    [({}, None), ({"bsdf": {"type": "lambertian", "reflectance": 0.0}}, 0.0)],
    ids=["noargs", "args"],
)
def test_sphere_construct_kernel_dict(modes_all, kwargs, expected_reflectance):
    sphere = SphereShape(**kwargs)
    mi_wrapper = check_scene_element(sphere, mi.Shape)

    if expected_reflectance is not None:
        assert mi_wrapper.parameters["bsdf.reflectance.value"] == expected_reflectance


def test_sphere_surface():
    # Check that parameters are correctly specified
    with ucc.override(length="km"):
        surface = SphereShape.surface(altitude=1.0, planet_radius=1.0)
        assert np.allclose(surface.center, [0, 0, 0] * ureg.km)
        assert np.allclose(surface.radius, 2.0 * ureg.km)


def test_sphere_atmosphere():
    # Check that parameters are correctly specified
    with ucc.override(length="km"):
        sphere = SphereShape.atmosphere(top=1.1, planet_radius=1.0)
        assert np.allclose(sphere.center, [0, 0, 0] * ureg.km)
        assert np.allclose(sphere.radius, 2.1 * ureg.km)


def test_sphere_contains():
    sphere = SphereShape(center=[1, 1, 1], radius=2.0)

    # Works with a single point
    assert sphere.contains([1.5, 1.5, 1.5])
    assert not sphere.contains([1.5, 2.0, -2.0])
    assert not sphere.contains([1.5, 1.5, 1.5] * ureg.km)

    # Works with multiple points
    assert np.all(sphere.contains([[1.5, 1.5, 1.5], [1.5, 2.0, -2.0]]) == [True, False])


def test_sphere_bbox():
    sphere = SphereShape(center=[1, 1, 1], radius=2.0)
    bbox = sphere.bbox
    np.testing.assert_array_equal(bbox.min.m_as(ureg.m), [-1, -1, -1])
    np.testing.assert_array_equal(bbox.max.m_as(ureg.m), [3, 3, 3])



"""
Global configuration components. The centric part is the
:class:`.EradiateConfig` class, whose defaults can be set using environment
variables. A single instance :data:`.config` is aliased in the top-level module
for convenience.
"""

import enum
import os.path
import pathlib

import attrs
import environ
from environ._environ_config import CNF_KEY, RAISE, _ConfigEntry
from environ.exceptions import ConfigError

from .frame import AzimuthConvention


def var(
    default=RAISE,
    converter=None,
    name=None,
    validator=None,
    help=None,
    on_setattr=None,
):
    """
    Reimplementation of `environ-config`'s :func:`var` with `on_setattr` support.

    Declare a configuration attribute on the body of `config`-decorated class.

    It will be attempted to be filled from an environment variable based on the
    prefix and *name*.

    Parameters
    ----------
    default
        Setting this to a value makes the config attribute optional.

    name : str
        Overwrite name detection with a string.  If not set, the name of the
        attribute is used.

    converter
        A callable that is run with the found value and its return value is
        used.  Please note that it is also run for default values.

    validator
        A callable that is run with the final value. See ``attrs``'s
        `chapter on validation <https://www.attrs.org/en/stable/init.html#validators>`_
        for details.
        You can also use any validator that is
        `shipped with attrs <https://www.attrs.org/en/stable/api.html#validators>`_.

    help : str
        A help string that is used by `generate_help`.

    on_setattr : callable or list of callables or None or attrs.setters.NO_OP, optional
        This argument is directly forwarded to :func:`attrs.field`, with the notable
        difference that the default behaviour executes converters and
        validators.
    """
    if on_setattr is None:
        on_setattr = attrs.setters.pipe(attrs.setters.convert, attrs.setters.validate)

    return attrs.field(
        default=default,
        metadata={CNF_KEY: _ConfigEntry(name, default, None, None, help)},
        converter=converter,
        validator=validator,
        on_setattr=on_setattr,
    )


class ProgressLevel(enum.IntEnum):
    """
    An enumeration defining valid progress levels.

    This is an integer enumeration, meaning that levels can be compared.
    """

    NONE = 0  #: No progress
    SPECTRAL_LOOP = enum.auto()  #: Up to spectral loop level progress
    KERNEL = enum.auto()  #: Up to kernel level progress


def format_help_dicts_rst(help_dicts, display_defaults=False):
    """
    Help dictionary formatter for environment variable help generation.
    """
    help_strs = []
    for help_dict in help_dicts:
        help_str = f".. envvar:: {help_dict['var_name']}\n\n"
        help_str += f"   *{'Required' if help_dict['required'] else 'Optional'}"

        if help_dict.get("default") and display_defaults:
            help_str += f", default={repr(help_dict['default'])}.* "
        else:
            help_str += "*. "

        if help_dict.get("help_str"):
            help_str += f"{help_dict['help_str']}"
        help_strs.append(help_str)

    return "\n\n".join(help_strs)


@environ.config(prefix="ERADIATE")
class EradiateConfig:
    """
    Global configuration for Eradiate.

    This class, instantiated once as the :data:`eradiate.config` attribute,
    contains global configuration parameters for Eradiate. It is initialized
    using :ref:`environment variables <sec-config-env_vars>` as defaults.

    See Also
    --------
    :data:`eradiate.config`,
    `the environ-config library <https://environ-config.readthedocs.io>`_
    """

    #: Path to the Eradiate source directory.
    source_dir = var(
        converter=lambda x: x if x is None else pathlib.Path(x).absolute(),
        help="Path to the Eradiate source directory. If it is not set, then the "
        "current setup is assumed to be a production installation of Eradiate.",
        default=None,
    )

    @source_dir.validator
    def _dir_validator(self, var, dir):
        if dir is None:
            # Import must be local and not use the lazy loader to avoid circular imports
            from .kernel._versions import kernel_installed

            is_installed, _ = kernel_installed()

            if not is_installed:
                raise ConfigError(
                    "Could not find a suitable production installation for the "
                    "Eradiate kernel. This is either because you are using Eradiate "
                    "in a production environment without having the eradiate-mitsuba "
                    "package installed, or because you are using Eradiate directly "
                    "from the sources. In the latter case, please make sure the "
                    "'ERADIATE_SOURCE_DIR' environment variable is correctly set to "
                    "the Eradiate installation directory. If you are using Eradiate "
                    "directly from the sources, you can alternatively source the "
                    "provided setpath.sh script. You can install the eradiate-mitsuba "
                    "package using 'pip install eradiate-mitsuba'."
                )
            return

        eradiate_init = dir / "src" / "eradiate" / "__init__.py"

        if not eradiate_init.is_file():
            raise ConfigError(
                f"While configuring Eradiate: could not find {eradiate_init} file. "
                "Please make sure the 'ERADIATE_SOURCE_DIR' environment variable is "
                "correctly set to the Eradiate installation directory. If you are "
                "using Eradiate directly from the sources, you can alternatively "
                "source the provided setpath.sh script. If you wish to use Eradiate "
                "in a production environment, you can install the eradiate-mitsuba "
                "package using 'pip install eradiate-mitsuba' and unset the "
                "'ERADIATE_SOURCE_DIR' environment variable."
            ) from FileNotFoundError(eradiate_init)

    #: URL where small data files are located (production use only)
    small_files_registry_url = var(
        default="https://raw.githubusercontent.com/eradiate/eradiate-data",
        converter=str,
        help="URL where small data files are located (production use only)",
    )

    #: Revision of the small files registry (production use only)
    small_files_registry_revision = var(
        default="master",
        converter=str,
        help="Revision of the small files registry (production use only)",
    )

    #: URL where large data files are located.
    data_store_url = var(
        default="http://eradiate.eu/data/store/",
        converter=str,
        help="URL where large data files are located.",
    )

    #: Path to the Eradiate download directory.
    download_dir = var(
        converter=lambda x: x if x is None else pathlib.Path(x).absolute(),
        help="Path to the Eradiate download directory.",
        default=None,
    )

    #: If ``True``, activate the offline mode. All online data stores
    #: will be disconnected.
    offline = environ.bool_var(
        default=False,
        help="If ``True``, activate the offline mode. All online data stores "
        "will be disconnected.",
    )

    #: An integer flag setting the level of progress display (see
    #: :class:`ProgressLevel`). Values are preferably set using strings
    #: (``["NONE", "SPECTRAL_LOOP", "KERNEL"]``). Only affects tqdm-based progress
    #: bars.
    progress = var(
        default="KERNEL",
        converter=lambda x: ProgressLevel[x.upper()]
        if isinstance(x, str)
        else ProgressLevel(x),
        help="An integer flag setting the level of progress display (see"
        ":class:`.ProgressLevel`). Values are preferably set using strings "
        '(``["NONE", "SPECTRAL_LOOP", "KERNEL"]``). Only affects tqdm-based '
        "progress bars.",
    )

    #: The convention applied when interpreting azimuth values as part
    #: of the specification of a direction (see :class:`.AzimuthConvention`).
    #: Values are preferably set using strings (*e.g.* ``"EAST_RIGHT"``,
    #: ``"NORTH_LEFT"``, etc.).
    azimuth_convention = var(
        default="EAST_RIGHT",
        converter=lambda x: AzimuthConvention[x.upper()] if isinstance(x, str) else x,
        validator=attrs.validators.instance_of(AzimuthConvention),
        help="The convention applied when interpreting azimuth values as part "
        "of the specification of a direction (see :class:`.AzimuthConvention`). "
        'Values are preferably set using strings (*e.g.* ``"EAST_RIGHT"``, '
        '``"NORTH_LEFT"``, etc.).',
    )


try:
    #: Global configuration object instance.
    #: See :class:`EradiateConfig`.
    config = EradiateConfig.from_environ()

except environ.exceptions.MissingEnvValueError as e:
    if "ERADIATE_SOURCE_DIR" in e.args:
        raise ConfigError(
            f"While configuring Eradiate: environment variable '{e}' is missing. "
            "Please set this variable to the absolute path to the Eradiate "
            "installation directory. If you are using Eradiate directly from the "
            "sources, you can alternatively source the provided setpath.sh script."
        ) from e
    else:
        raise e



import os.path
from datetime import datetime
from pathlib import Path

import pytest
import xarray as xr

from eradiate.data import SafeOnlineDataStore
from eradiate.exceptions import DataError

TEST_STORE = "http://eradiate.eu/data/store/stable"
TEST_FILE = Path("tests/data/online/registered_dataset.nc")


def test_online_data_store_registry(tmpdir):
    # Using an empty storage directory, a SafeOnlineDataStore instance can be created
    store = SafeOnlineDataStore(base_url=TEST_STORE, path=tmpdir)

    # The registry file is downloaded
    registry_filename = store.registry_fetch()
    assert registry_filename.is_file()

    # Reloading the registry works as expected
    store.registry["foo"] = "bar"
    store.registry_reload()
    assert "foo" not in store.registry

    # We can force a registry download
    last_state = store.registry.copy()
    last_modified = datetime.fromtimestamp(registry_filename.stat().st_mtime)
    store.registry_reload(delete=True)
    # The registry contents are the same but the file is more recent
    assert store.registry == last_state
    assert store.registry is not last_state
    assert datetime.fromtimestamp(registry_filename.stat().st_mtime) > last_modified


def test_oneline_data_store_is_registered(tmpdir):
    store = SafeOnlineDataStore(base_url=TEST_STORE, path=tmpdir)

    # Default behaviour allows matching compressed files
    filename = store.is_registered(TEST_FILE)
    assert str(filename) == TEST_FILE.as_posix() + ".gz"

    # The test file is actually gzip-compressed and will therefore not match a
    # registry entry if allow_compressed is set to False
    with pytest.raises(ValueError):
        store.is_registered(TEST_FILE, allow_compressed=False)


def test_online_data_store_fetch(tmpdir):
    store = SafeOnlineDataStore(base_url=TEST_STORE, path=tmpdir)
    filename = os.path.join(tmpdir, TEST_FILE)

    # Upon calling fetch(), gzip detection is successful: the gz file is downloaded
    assert not os.path.isfile(filename + ".gz")
    store.fetch(TEST_FILE)
    assert os.path.isfile(filename + ".gz")

    # The gz file is also decompressed
    assert os.path.isfile(filename)
    with xr.open_dataset(store.fetch(TEST_FILE)) as ds:
        assert isinstance(ds, xr.Dataset)

    # A direct and blind download is attempted if the file is not registered
    # In this specific case, it will fail because the file does not exist on the
    # server
    with pytest.raises(DataError):
        store.fetch("foo")


def test_online_data_store_purge(tmpdir):
    # We start with an empty directory
    assert not any(os.scandir(tmpdir))

    # We create a data store and fetch a file:
    # we expect the directory to contain exactly 2 elements (the registry file
    # and the directory containing the fetched file)
    store = SafeOnlineDataStore(base_url=TEST_STORE, path=tmpdir)
    store.fetch(TEST_FILE)
    assert len(list(os.scandir(tmpdir))) == 2
    assert store.registry_path.is_file()
    assert (store.path / TEST_FILE).is_file()

    # We purge the directory but keep registered files
    os.makedirs(store.path / "foo")  # This empty directory should also be cleaned up
    store.purge(keep="registered")
    assert len(list(os.scandir(tmpdir))) == 2  # We still have registry.txt and a subdir
    assert store.registry_path.is_file()
    assert not (store.path / TEST_FILE).is_file()  # The decompressed file is gone
    assert (
        store.path / (str(TEST_FILE) + ".gz")
    ).is_file()  # The compressed file is still here

    # Purge everything: all files are removed
    store.purge()
    assert len(list(os.scandir(tmpdir))) == 0



import numpy as np
import pytest
import xarray as xr

import eradiate.test_tools.regression as tt


def test_instantiate():

    # instantiate the test with reasonable defaults
    assert tt.RMSETest(
        name="rmse",
        archive_dir="tests/",
        value=xr.Dataset(),
        reference=xr.Dataset(),
        threshold=0.05,
    )

    assert tt.Chi2Test(
        name="chi2",
        archive_dir="tests/",
        value=xr.Dataset(),
        reference=xr.Dataset(),
        threshold=0.05,
    )

    # assert all arguments except reference are needed
    with pytest.raises(TypeError):
        assert tt.Chi2Test(
            archive_dir="tests/",
            value=xr.Dataset(),
            reference=xr.Dataset(),
            threshold=0.05,
        )

    with pytest.raises(TypeError):
        tt.Chi2Test(
            name="chi2",
            value=xr.Dataset(),
            reference=xr.Dataset(),
            threshold=0.05,
        )

    with pytest.raises(TypeError):
        tt.Chi2Test(
            name="chi2",
            archive_dir="tests/",
            reference=xr.Dataset(),
            threshold=0.05,
        )

    with pytest.raises(TypeError):
        tt.Chi2Test(
            name="chi2",
            archive_dir="tests/",
            value=xr.Dataset(),
            reference=xr.Dataset(),
        )

    assert tt.Chi2Test(
        name="chi2",
        archive_dir="tests/",
        value=xr.Dataset(),
        threshold=0.05,
    )


def test_reference_converter(tmp_path):
    # test proper handling of missing and unreadable reference

    # file does not exist
    test = tt.Chi2Test(
        name="chi2",
        archive_dir="tests/",
        value=xr.Dataset(),
        threshold=0.05,
        reference="./this/file/doesnot.exist",
    )

    assert test.reference is None

    # wrong file type
    tempfile = tmp_path / "hello.txt"
    tempfile.write_text("test")

    with pytest.raises(ValueError):
        tt.Chi2Test(
            name="chi2",
            archive_dir="tests/",
            value=xr.Dataset(),
            threshold=0.05,
            reference=tempfile,
        )

    # wrong data type
    with pytest.raises(ValueError):
        tt.Chi2Test(
            name="chi2",
            archive_dir="tests/",
            value=xr.Dataset(),
            threshold=0.05,
            reference=np.zeros(25),
        )


def test_rmse_evaluate():
    # test the computation of the RMSE value from given data.
    # we give the dataset some wrong data fields to ensure the right
    # data is used

    result = np.random.rand(50)
    ref = np.random.rand(50)

    result_da = xr.DataArray(result)
    ref_da = xr.DataArray(ref)

    result_ds = xr.Dataset(
        data_vars={
            "brf": result_da,
            "stuff": result_da * 0.1,
            "wrong": result_da * 123.0,
        }
    )
    ref_ds = xr.Dataset(
        data_vars={"brf": ref_da, "stuff": ref_da * 0.2, "wrong": ref_da * 321.0}
    )

    rmse_ref = np.linalg.norm(result - ref) / np.sqrt(len(ref))

    test = tt.RMSETest(
        name="rmse",
        value=result_ds,
        reference=ref_ds,
        archive_dir="tests/",
        threshold=0.05,
    )

    _, rmse = test._evaluate()

    assert rmse == rmse_ref


def test_chi2_evaluate(mode_mono):
    # test the computation of the Chi squared value from given data.
    # we give the dataset some wrong data fields to ensure the right
    # data is used

    import mitsuba as mi

    result_np = np.random.rand(50)
    ref_np = np.random.rand(50)

    histo_bins = np.linspace(ref_np.min(), ref_np.max(), 20)
    histo_ref = np.histogram(ref_np, histo_bins)[0]
    histo_res = np.histogram(result_np, histo_bins)[0]

    # sorting both histograms following the ascending frequencies in
    # the reference. Algorithm from:
    # https://stackoverflow.com/questions/9764298/how-to-sort-two-lists-which-reference-each-other-in-the-exact-same-way
    histo_ref_sorted, histo_res_sorted = zip(
        *sorted(zip(histo_ref, histo_res), key=lambda x: x[0])
    )

    from mitsuba.python.math import rlgamma

    chi2val, dof, pooled_in, pooled_out = mi.math.chi2(
        histo_res_sorted, histo_ref_sorted, 5
    )
    p_value_ref = 1.0 - rlgamma(dof / 2.0, chi2val / 2.0)

    result_da = xr.DataArray(result_np)
    ref_da = xr.DataArray(ref_np)

    result_ds = xr.Dataset(
        data_vars={
            "brf": result_da,
            "stuff": result_da * 0.1,
            "wrong": result_da * 123.0,
        }
    )
    ref_ds = xr.Dataset(
        data_vars={"brf": ref_da, "stuff": ref_da * 0.2, "wrong": ref_da * 321.0}
    )

    test = tt.Chi2Test(
        name="chi2",
        value=result_ds,
        reference=ref_ds,
        archive_dir="tests/",
        threshold=0.05,
    )

    _, p_value = test._evaluate()

    assert p_value == p_value_ref



import pytest

import eradiate
from eradiate.data import MultiDataStore, SafeDirectoryDataStore, SafeOnlineDataStore
from eradiate.exceptions import DataError

TEST_STORE = "http://eradiate.eu/data/store/stable"
TEST_FILE_SUBMODULE_REGISTERED = "tests/data/git/registered_dataset.nc"
TEST_FILE_SUBMODULE_UNREGISTERED = "tests/data/git/unregistered_dataset.nc"
TEST_FILE_ONLINE_REGISTERED = "tests/data/online/registered_dataset.nc"
TEST_FILE_ONLINE_UNREGISTERED = "tests/data/online/unregistered_dataset.nc"


def test_multi_data_store(tmpdir):
    # Initialise test data store
    data_store = MultiDataStore(
        [
            (
                "1",
                SafeDirectoryDataStore(
                    path=eradiate.config.source_dir / "resources/data"
                ),
            ),
            (
                "2",
                SafeOnlineDataStore(base_url=TEST_STORE, path=tmpdir),
            ),
        ]
    )

    # Fetch files from both data stores
    assert data_store.fetch(TEST_FILE_SUBMODULE_REGISTERED)
    assert data_store.fetch(TEST_FILE_ONLINE_REGISTERED)

    # Unregistered data will go through only if they are available from the
    # online store
    with pytest.raises(DataError):
        data_store.fetch(TEST_FILE_ONLINE_UNREGISTERED)
    with pytest.raises(DataError):
        data_store.fetch(TEST_FILE_SUBMODULE_UNREGISTERED)



import mitsuba as mi
import pytest

from eradiate.scenes.phase import HenyeyGreensteinPhaseFunction
from eradiate.test_tools.types import check_scene_element


@pytest.mark.parametrize("kwargs", [{}, {"g": 0.25}], ids=["noargs", "args"])
def test_hg(modes_all_double, kwargs):
    phase = HenyeyGreensteinPhaseFunction(**kwargs)
    check_scene_element(phase, mi.PhaseFunction)



import mitsuba as mi
import numpy as np
import pytest
import xarray as xr

import eradiate
from eradiate import unit_registry as ureg
from eradiate.constants import EARTH_RADIUS
from eradiate.experiments import AtmosphereExperiment
from eradiate.scenes.atmosphere import HomogeneousAtmosphere
from eradiate.scenes.measure import MultiDistantMeasure
from eradiate.scenes.shapes import RectangleShape
from eradiate.scenes.spectra import MultiDeltaSpectrum
from eradiate.test_tools.types import check_scene_element
from eradiate.test_tools.util import skipif_data_not_found
from eradiate.units import to_quantity


def test_atmosphere_experiment_construct_default(modes_all_double):
    """
    AtmosphereExperiment initializes successfully with default parameters in
    all modes.
    """
    exp = AtmosphereExperiment()

    # Check that the atmosphere's geometry is overridden by the experiment's
    assert exp.atmosphere.geometry is exp.geometry


def test_atmosphere_experiment_extra_objects(mode_mono):
    """
    Extra objects can be added to the scene.
    """
    exp = AtmosphereExperiment(
        extra_objects={
            "reference_surface": {
                "factory": "shape",
                "type": "rectangle",
                "id": "reference_surface",
                "bsdf": {"type": "lambertian"},
            }
        }
    )
    assert isinstance(exp.extra_objects["reference_surface"], RectangleShape)
    mi_wrapper = check_scene_element(exp.scene, mi.Scene, ctx=exp.context_init)
    assert mi_wrapper.obj.shapes()[0].id() == "reference_surface"
    assert "reference_surface.bsdf.reflectance.value" in mi_wrapper.parameters.keys()


def test_atmosphere_experiment_construct_measures(modes_all_double):
    """
    A variety of measure specifications are acceptable.
    """
    # Init with a single measure (not wrapped in a sequence)
    assert AtmosphereExperiment(measures=MultiDistantMeasure())

    # Init from a dict-based measure spec
    # -- Correctly wrapped in a sequence
    assert AtmosphereExperiment(measures=[{"type": "distant"}])
    # -- Not wrapped in a sequence
    assert AtmosphereExperiment(measures={"type": "distant"})


@pytest.mark.parametrize(
    "geometry, expected",
    [
        ("plane_parallel", [0, 0, 0] * ureg.m),
        ("spherical_shell", [0, 0, EARTH_RADIUS.m] * EARTH_RADIUS.u),
    ],
    ids=["plane_parallel", "spherical_shell"],
)
def test_atmosphere_experiment_construct_normalize_measures(
    mode_mono, geometry, expected
):
    """
    Default measure target is set to ground level.
    """
    exp = AtmosphereExperiment(geometry=geometry)
    assert np.allclose(exp.measures[0].target.xyz, expected)


def test_atmosphere_experiment_kernel_dict(mode_mono):
    """
    Test non-trivial kernel dict generation behaviour.
    """
    skipif_data_not_found(
        "spectra/absorption/us76_u86_4/us76_u86_4-spectra-18000_19000.nc"
    )

    # With an atmosphere
    exp = AtmosphereExperiment(
        geometry={"type": "plane_parallel", "width": 42.0, "width_units": "km"},
        atmosphere=HomogeneousAtmosphere(),
        measures={"type": "distant"},
    )
    mi_wrapper = check_scene_element(
        exp.scene, mi.Scene, drop_parameters=False
    )  # Do not drop untracked parameters: we want to check the `to_world` transform
    # -- Surface width is inherited from geometry
    assert np.allclose(
        mi_wrapper.parameters["surface_shape.to_world"].matrix,
        mi.ScalarTransform4f.scale([21000, 21000, 1]).matrix,
    )
    # -- Atmosphere is part of the scene
    assert "shape_atmosphere" in set(shape.id() for shape in mi_wrapper.obj.shapes())
    # -- Measure gets no external medium assigned
    assert all(sensor.medium() is None for sensor in mi_wrapper.obj.sensors())

    # Without an atmosphere
    exp = AtmosphereExperiment(
        geometry="plane_parallel",
        atmosphere=None,
        surface={"type": "lambertian"},
        measures=[
            {"type": "distant", "id": "distant_measure"},
            {"type": "radiancemeter", "origin": [1, 0, 0], "id": "radiancemeter"},
        ],
    )
    # -- Surface width has default value
    mi_wrapper = check_scene_element(exp.scene, mi.Scene, drop_parameters=False)
    assert np.allclose(
        mi_wrapper.parameters["surface_shape.to_world"].matrix,
        mi.ScalarTransform4f.scale([5e8, 5e8, 1]).matrix,
    )
    # -- Atmosphere is not part of the scene
    assert "shape_atmosphere" not in set(
        shape.id() for shape in mi_wrapper.obj.shapes()
    )
    # -- Measures get no external medium assigned
    assert all(sensor.medium() is None for sensor in mi_wrapper.obj.sensors())


def test_atmosphere_experiment_real_life(
    mode_mono,
    us_standard_mono,
):
    # Construct with typical parameters
    exp = AtmosphereExperiment(
        surface={"type": "rpv"},
        atmosphere={
            "type": "heterogeneous",
            "molecular_atmosphere": us_standard_mono,
        },
        illumination={"type": "directional", "zenith": 45.0},
        measures=[
            {"type": "distant", "id": "distant_measure"},
            {"type": "radiancemeter", "origin": [1, 0, 0], "id": "radiancemeter"},
        ],
    )
    mi_wrapper = check_scene_element(exp.scene, mi.Scene, ctx=exp.context_init)

    # -- Distant measures get no external medium
    mi_sensors = {sensor.id(): sensor for sensor in mi_wrapper.obj.sensors()}
    assert mi_sensors["distant_measure"].medium() is None

    # -- Radiancemeter inside the atmosphere must have a medium assigned
    assert mi_sensors["radiancemeter"].medium().id() == "medium_atmosphere"


def test_atmosphere_experiment_run_basic(
    modes_all_double,
    us_standard_mono,
    us_standard_ckd_550nm,
):
    if eradiate.mode().is_mono:
        molecular_atmosphere = us_standard_mono
    else:
        molecular_atmosphere = us_standard_ckd_550nm

    exp = AtmosphereExperiment(
        atmosphere={
            "type": "heterogeneous",
            "molecular_atmosphere": molecular_atmosphere,
            "particle_layers": [{"type": "particle_layer"}],
        },
        surface={"type": "lambertian"},
        measures={
            "type": "distant",
            "id": "distant_measure",
            "srf": MultiDeltaSpectrum(wavelengths=550.0 * ureg.nm),
        },
    )
    assert isinstance(eradiate.run(exp, spp=100), xr.Dataset)


def test_atmosphere_experiment_run_detailed(
    modes_all_double,
    us_standard_mono,
    us_standard_ckd_550nm,
):
    """
    Test for correctness of the result dataset generated by AtmosphereExperiment.
    """
    if eradiate.mode().is_mono:
        molecular_atmosphere = us_standard_mono
    else:
        molecular_atmosphere = us_standard_ckd_550nm

    # Create simple scene
    exp = AtmosphereExperiment(
        measures=[
            {
                "type": "hemispherical_distant",
                "id": "toa_hsphere",
                "film_resolution": (32, 32),
                "spp": 100,
                "srf": MultiDeltaSpectrum(wavelengths=550.0 * ureg.nm),
            },
        ],
        atmosphere=molecular_atmosphere,
    )

    # Run RT simulation
    results = eradiate.run(exp)

    # Check result dataset structure
    # Post-processing creates expected variables ...
    expected = {"brdf", "brf", "irradiance", "radiance"}
    assert set(results.data_vars) == expected

    # ... dimensions ...
    assert set(results["radiance"].dims) == {"saa", "sza", "w", "x_index", "y_index"}
    assert set(results["irradiance"].dims) == {"saa", "sza", "w"}

    # ... and other coordinates
    expected_coords = {
        "saa",
        "sza",
        "vaa",
        "vza",
        "w",
        "x",
        "x_index",
        "y",
        "y_index",
    }
    if eradiate.mode().is_ckd:
        expected_coords |= {"bin_wmin", "bin_wmax"}
    assert set(results["radiance"].coords) == expected_coords

    expected_coords = {"saa", "sza", "w"}
    if eradiate.mode().is_ckd:
        expected_coords |= {"bin_wmin", "bin_wmax"}
    assert set(results["irradiance"].coords) == expected_coords

    # We just check that we record something as expected
    assert np.all(results["radiance"].data > 0.0)


def test_atmosphere_experiment_custom_atmosphere(mode_ckd, cams_lybia4_ckd_550nm):
    """
    Test that the AtmosphereExperiment can deal with a custom atmosphere (CAMS).
    """
    # Create a geometry compatible with the molecular atmosphere
    zgrid = to_quantity(cams_lybia4_ckd_550nm["thermoprops"].z)
    geometry = {
        "type": "spherical_shell",
        "zgrid": zgrid,
        "toa_altitude": zgrid[-1],
        "ground_altitude": zgrid[0],
    }

    # Create simple scene
    exp = AtmosphereExperiment(
        measures=[
            {
                "type": "hemispherical_distant",
                "id": "toa_hsphere",
                "film_resolution": (32, 32),
                "spp": 100,
                "srf": MultiDeltaSpectrum(wavelengths=550.0 * ureg.nm),
            },
        ],
        atmosphere=cams_lybia4_ckd_550nm,
        geometry=geometry,
    )

    # Run RT simulation
    results = eradiate.run(exp, spp=100)

    # We just check that we record something as expected
    assert np.all(results["radiance"].data > 0.0)



import os


def main():
    # Check if conda env is active
    conda_default_env = os.environ.get("CONDA_DEFAULT_ENV")

    if not conda_default_env or conda_default_env == "base":
        print("Error: Activate a Conda environment other than 'base'")
        exit(1)


if __name__ == "__main__":
    main()



"""A series of basic one-dimensional test cases."""

from contextlib import nullcontext

import mitsuba as mi
import numpy as np
import pytest

import eradiate
from eradiate import KernelContext
from eradiate import unit_registry as ureg
from eradiate.kernel import mi_render
from eradiate.scenes.bsdfs import LambertianBSDF
from eradiate.scenes.core import Scene
from eradiate.scenes.illumination import (
    AstroObjectIllumination,
    ConstantIllumination,
    DirectionalIllumination,
)
from eradiate.scenes.integrators import PathIntegrator
from eradiate.scenes.measure import MultiDistantMeasure
from eradiate.scenes.shapes import RectangleShape
from eradiate.scenes.surface import BasicSurface
from eradiate.test_tools.types import check_scene_element


@pytest.mark.parametrize(
    "illumination, spp",
    [
        ("directional", 1),
        ("constant", 5e5),
        # Deactivated for now (inexplicably fails when run as part of full test suite)
        # ("astro_object", 5e5),
    ],
)
@pytest.mark.parametrize("li", [0.1, 1.0, 10.0])
@pytest.mark.slow
def test_radiometric_accuracy(modes_all_mono, illumination, spp, li, ert_seed_state):
    r"""
    Radiometric check (``path``)
    ============================

    This simple test case compares simulated leaving radiance at a Lambertian
    surface with theoretical values.

    Rationale
    ---------

    * Geometry: a square surface with unit size and a Lambertian BRDF with
      reflectance :math:`\rho = 0.5`.
    * Illumination:

      * ``directional``: a directional light source at the zenith with
        radiance :math:`L_\mathrm{i} \in [0.1, 1, 10]`.
      * ``constant``: an isotropic illumination with radiance
        :math:`L_\mathrm{i} \in [0.1, 1, 10]`.

    * Sensor: A ``distant`` sensor targeting (0, 0, 0).


    Expected behaviour
    ------------------

    The computed solution is equal to the theoretical solution (relative
    tolerance of 0.1%).

    Theoretical solutions:

    * ``directional``: :math:`L_\mathrm{o} = \frac{\rho L_\mathrm{i}}{\pi}`
    * ``constant``: :math:`L_\mathrm{o} = \rho L_\mathrm{i}`
    """

    # Basic configuration
    vza = np.linspace(0, 80, 10)
    rho = 0.5

    # Ignore warning in single precision
    with pytest.warns(UserWarning, match="the selected mode is single-precision") if (
        eradiate.mode().is_single_precision and spp > 100000
    ) else nullcontext():
        measure = MultiDistantMeasure.hplane(
            zeniths=vza, azimuth=0.0, target=[0, 0, 0], spp=spp
        )

    objects = {
        "surface": BasicSurface(
            bsdf=LambertianBSDF(reflectance=rho),
            shape=RectangleShape(edges=2.0 * ureg.m),
        ),
        "measure": measure,
        "integrator": PathIntegrator(),
    }

    if illumination == "directional":
        objects["illumination"] = DirectionalIllumination(zenith=0.0, irradiance=li)
        theoretical_solution = np.full_like(vza, rho * li / np.pi)
        rtol = 1e-3

    elif illumination == "constant":
        objects["illumination"] = ConstantIllumination(radiance=li)
        theoretical_solution = np.full_like(vza, rho * li)
        rtol = 5e-3

    # Deactivated for now (see parametrization)
    elif illumination == "astro_object":
        objects["illumination"] = AstroObjectIllumination(
            zenith=0.0, irradiance=li, angular_diameter=0.03
        )
        theoretical_solution = np.full_like(vza, rho * li / np.pi)
        # The angular diameter is not taken into account in the theoretical solution
        rtol = 1e-2

    else:
        raise ValueError(f"unsupported illumination '{illumination}'")

    scene = Scene(objects=objects)
    mi_wrapper = check_scene_element(scene, mi.Scene)

    result = np.squeeze(mi_render(mi_wrapper, ctxs=[KernelContext()])[550.0]["measure"])
    np.testing.assert_allclose(result, theoretical_solution, rtol=rtol)



"""A series of test cases the assert the plausibility of Eradiate's computation."""

import os

import matplotlib.pyplot as plt
import numpy as np
import pytest

import eradiate
from eradiate import unit_registry as ureg
from eradiate.plot import remove_xylabels


def ensure_dir(path):
    if not os.path.isdir(path):
        os.makedirs(path)


@pytest.mark.parametrize("atmosphere", ["none", "homogeneous"])
@pytest.mark.parametrize("surface", ["lambertian", "rpv"])
@pytest.mark.slow
def test_symmetry_zenith(mode_mono_double, surface, atmosphere, artefact_dir):
    r"""
    Result symmetry
    ===============

    This test case checks for result symmetry when using the
    ``AtmosphereExperiment`` workflow.

    Rationale
    ---------

    We use a scene with rotational symmetry where the directional illumination
    targets the zenith direction.

    * Geometry: Either a Lambertian or RPV surface.
    * Illumination: Directional illumination from the zenith (default
      irradiance).
    * Atmosphere: Either no atmosphere or a purely Rayleigh scattering
      homogeneous atmosphere.
    * Sensor: Distant measure covering a plane (17 angular points, :math:`10^6`
      SPP) and targeting (0, 0, 0).
      We take an odd number of angular points to ensure that the special value
      :math:`\theta = 0°` is included.

    Expected behaviour
    ------------------

    The results on either side of the zenith must agree within a relative
    tolerance of 5e-3.

    Results
    -------

    For each of the four combinations of atmosphere and surface, two plots are
    shown. On the left of each pair, the recorded leaving radiance is plotted
    against the viewing zenith angle on top of its mirror image for a visual
    inspection of the result symmetry. On the right of each pair the absolute
    difference between these two curves is shown, for a more quantitative
    visualization.

    .. image:: generated/plots/test_symmetry_zenith_lambertian_none.png
       :width: 45%
    .. image:: generated/plots/test_symmetry_zenith_lambertian_homogeneous.png
       :width: 45%

    .. image:: generated/plots/test_symmetry_zenith_rpv_none.png
       :width: 45%
    .. image:: generated/plots/test_symmetry_zenith_rpv_homogeneous.png
       :width: 45%
    """
    spp = int(1e6)
    n_vza = 17

    # Run simulation
    exp = eradiate.experiments.AtmosphereExperiment(
        geometry={"type": "plane_parallel", "toa_altitude": 1.0e2 * ureg.km},
        illumination={"type": "directional", "zenith": 0.0, "azimuth": 0.0},
        measures={
            "type": "distant",
            "id": "toa_pplane",
            "construct": "hplane",
            "zeniths": np.linspace(-89, 89, n_vza),
            "azimuth": 0.0,
            "spp": spp,
        },
        surface={
            "lambertian": {"type": "lambertian", "reflectance": 0.5},
            "rpv": {"type": "rpv", "rho_0": 0.183, "k": 0.780, "g": -0.1},
        }[surface],
        atmosphere={
            "none": None,
            "homogeneous": {
                "type": "homogeneous",
                "sigma_s": 1.0e-2 * ureg.km**-1,
            },
        }[atmosphere],
    )
    results = eradiate.run(exp)

    # Post-process results
    radiance = results["radiance"].squeeze().values
    results["diff"] = (
        results["radiance"].dims,
        (radiance - radiance[::-1]).reshape(results["radiance"].shape),
    )
    results["radiance_reversed"] = (
        results["radiance"].dims,
        radiance[::-1].reshape(results["radiance"].shape),
    )

    # Plot results
    fig, [ax1, ax2] = plt.subplots(1, 2, figsize=(5, 2.5))
    results["radiance"].plot(ax=ax1, x="vza", marker=".", ls="--")
    results["radiance_reversed"].plot(ax=ax1, x="vza", marker=".", ls="--")
    results["diff"].plot(ax=ax2, x="vza", marker=".", ls="--")
    remove_xylabels(from_=[ax1, ax2])
    ax2.yaxis.tick_right()
    ax1.set_title("")
    ax2.set_title("")
    plt.suptitle(f"Surface: {surface}, Atmosphere: {str(atmosphere)}")
    plt.tight_layout()

    filename = f"test_symmetry_zenith_{surface}_{str(atmosphere)}.png"
    ensure_dir(os.path.join(artefact_dir, "plots"))
    fname_plot = os.path.join(artefact_dir, "plots", filename)

    fig.savefig(fname_plot, dpi=200)
    plt.close()

    # Check symmetry
    assert np.allclose(
        results["radiance"].squeeze(),
        results["radiance"].squeeze().loc[::-1].data,
        rtol=5e-3,
    )



"""
Utilities for the Eradiate test suite.
"""
from __future__ import annotations

import os
import typing as t
from pathlib import Path

import matplotlib.pyplot as plt
import pytest

from ..data import data_store
from ..exceptions import DataError
from ..typing import PathLike


def skipif_data_not_found(path: PathLike, action: t.Callable | None = None) -> None:
    """
    During a Pytest session, skip the current test if the referenced dataset
    cannot be fetched from the data store.

    Parameters
    ----------
    path : path-like
        Path to the required data file (in the data store).

    action : callable, optional
        An optional callable with no arguments which performs an action prior to
        skipping the test (*e.g.* output an artefact placeholder).
    """
    try:
        data_store.fetch(path)
    except DataError:
        if action is not None:
            action()

        pytest.skip(f"Could not find dataset '{path}' in the data store.")


def missing_artefact(filename: PathLike) -> None:
    """
    Create a placeholder file for a missing artefact.

    Parameters
    ----------
    filename : path-like
        Path to the created placeholder.

    Raises
    ------
    ValueError
        Required file type is not supported.

    Notes
    -----
    The following types are currently supported:

    * PNG
    """
    filename = Path(filename)

    if filename.suffix == ".png":
        plt.figure()
        plt.text(0.5, 0.5, "Placeholder", fontsize="xx-large", ha="center")
        plt.gca().set_aspect(1)
        os.makedirs(filename.parent, exist_ok=True)
        plt.savefig(filename)
        plt.close()

    else:
        raise ValueError(f"unsupported file extension {filename.suffix}")



import numpy as np
import pytest

import eradiate
from eradiate.data import data_store
from eradiate.experiments import CanopyExperiment
from eradiate.test_tools.regression import Chi2Test
from eradiate.units import unit_registry as ureg


@pytest.mark.regression
def test_het04a1_brfpp(mode_mono_double, artefact_dir, session_timestamp):
    r"""
    Real zoom in (HET04a1) regression test
    ======================================

    This is a regression test, which compares the simulation results of the
    current branch to an older reference version.

    Rationale
    ---------

    This test case implements a basic canopy scene:

    * Surface with lambertian reflectance
    * No atmosphere
    * Three dimensional canopy

    Parameters

    * Surface: Square surface with labmertian BSDF with :math:`r = 0.15`
    * Canopy:

      * Floating spheres made up of disks with bilambertian bsdf model.
        Leaf reflectance is 0.49, transmittance is 0.41.
      * Floating cylinders made up of disks with bilambertian bsdf model.
        Leaf reflectance is 0.45, transmittance is 0.43.

      Disk, sphere and cylinder positioning follow the HET04 scenario of the
      RAMI-3 benchmark.
    * Illumination: Directional illumination with a zenith angle :math:`\theta = 20°`
    * Sensor: Distant reflectance measure, covering a plane, (76 angular points,
      10000 samples per pixel). This test implements the variant a1 of the HET04
      scenario, in which the entire scene is targeted by the sensor.

    Expected behaviour
    ------------------

    This test uses the Chi-squared criterion with a threshold of 0.05.

    """

    leaf_spec_path_sph = data_store.fetch(
        "tests/regression_test_specifications/het04/het04_sph.def"
    )
    leaf_pos_path_sph = data_store.fetch(
        "tests/regression_test_specifications/het04/het04_sph_positions.def"
    )
    leaf_spec_path_cyl = data_store.fetch(
        "tests/regression_test_specifications/het04/het04_cyl.def"
    )
    leaf_pos_path_cyl = data_store.fetch(
        "tests/regression_test_specifications/het04/het04_cyl_positions.def"
    )

    exp = CanopyExperiment(
        canopy={
            "type": "discrete_canopy",
            "instanced_canopy_elements": [
                {
                    "construct": "from_file",
                    "filename": leaf_pos_path_sph,
                    "canopy_element": {
                        "type": "leaf_cloud",
                        "construct": "from_file",
                        "filename": leaf_spec_path_sph,
                        "leaf_reflectance": 0.49,
                        "leaf_transmittance": 0.41,
                        "id": "spherical_leaf_cloud",
                    },
                },
                {
                    "construct": "from_file",
                    "filename": leaf_pos_path_cyl,
                    "canopy_element": {
                        "type": "leaf_cloud",
                        "construct": "from_file",
                        "filename": leaf_spec_path_cyl,
                        "leaf_reflectance": 0.45,
                        "leaf_transmittance": 0.3,
                        "id": "cylindrical_leaf_cloud",
                    },
                },
            ],
            "size": [270, 270, 15] * ureg.m,
        },
        surface={"type": "lambertian", "reflectance": 0.15},
        padding=5,
        measures=[
            {
                "type": "mdistant",
                "construct": "hplane",
                "spp": 20000,
                "azimuth": 180 * ureg.deg,
                "zeniths": np.arange(-75, 75.01, 2) * ureg.deg,
                "target": {
                    "type": "rectangle",
                    "xmin": -135 * ureg.m,
                    "xmax": 135 * ureg.m,
                    "ymin": -135 * ureg.m,
                    "ymax": 135 * ureg.m,
                    "z": 15 * ureg.m,
                },
            }
        ],
        illumination={
            "type": "directional",
            "zenith": 20 * ureg.deg,
            "azimuth": 0.0 * ureg.deg,
            "irradiance": 20.0,
        },
        integrator={"type": "path", "max_depth": -1},
    )

    result = eradiate.run(exp)

    test = Chi2Test(
        name=f"{session_timestamp:%Y%m%d-%H%M%S}-het04.nc",
        value=result,
        reference="tests/regression_test_references/het04_brfpp_ref.nc",
        threshold=0.05,
        archive_dir=artefact_dir,
    )

    assert test.run()
